#############################################################################
# main.tcl Copyright 2004 r0lZ
# PgcEdit main window GUI and procedures.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::main {

#############################################################################
## Procedure:  about
## Display the about message
proc about {} {
	set txt "Version $::program_version ($::release_date)"
	set txt "$txt\nCopyright r0lZ, 2004"

	set txt "$txt\n\nPgcEdit is free software (donnationware)"
	set txt "$txt\ndistributed under the GPL license (GNU General Public License)"
	set txt "$txt\nUse PgcEdit at your own risk!"

	set txt "$txt\n\nIf you like PgcEdit, please make a donnation, to support future development."

	set txt "$txt\n\nUsing Tcl/Tk version $::tcl_patchLevel"
	if {[info exists ::freewrap::progname]} {
		set txt "$txt\nWrapped with FreeWrap $::freewrap::patchLevel"
	}
	set txt "$txt\nRunning on $::tcl_platform(os) $::tcl_platform(osVersion)"

	set txt "$txt\n\nPgcEdit home: http://home.tiscali.be/debie.roland/pgcedit/"
	set txt "$txt\nMirror hosted by VideoHelp: http://www.videohelp.com/~r0lZ/pgcedit/"
	set txt "$txt\nOfficial PgcEdit help and contact forum hosted by Doom9,"
	set txt "$txt\n(http://forum.doom9.org) in category DVD and miniDVD > IFO/VOB Editors."

	set txt "$txt\n\nBased on DVD Informations by MPUCoder: http://www.mpucoder.com/dvd/"

	set txt "$txt\n\nThanks to everybody at Doom9 forum!"
	set txt "$txt\nSpecial thanks to Jsoto and Jeanl for their support, to Sweetness for the toolbar icons,"
	set txt "$txt\nand to 2COOL, $$$ and Jeanl for the beta tests and ideas."
	if {$::tcl_platform(platform) == "windows"} {
		set txt "$txt\nCredits to www.dvd-replica.com for online VM commands documentation."
	}

	toplevel .about
	wm withdraw .about
	wm title .about "About PgcEdit"
	wm transient .about .
	wm overrideredirect .about 1
	wm resizable .about 0 0

	frame .about.a -borderwidth 1 -relief sunken
	pack  .about.a -side top -fill both -expand 1
	frame .about.a.b -borderwidth 6 -relief ridge
	pack  .about.a.b -side top -fill both -expand 1 -anchor center
	frame .about.a.b.pad1 -height 20
	pack  .about.a.b.pad1 -side top -fill y -expand 1
	label .about.a.b.lab1 -text "PgcEdit" -bd 2 -relief sunken \
			-padx 12 -pady 10 -background #EEEEEE
	pack  .about.a.b.lab1 -side top -padx 20 -anchor center -pady 8
	.about.a.b.lab1 configure -font "\"[lindex [.about.a.b.lab1 cget -font] 0]\" 14"
	label .about.a.b.lab2 -text $txt
	pack  .about.a.b.lab2 -side top -padx 20 -anchor center
	frame .about.a.b.pad2 -height 20
	pack  .about.a.b.pad2 -side top -fill y -expand 1

	update
	if {[winfo reqwidth .about] > [winfo reqheight .about]} {
		set size [winfo reqwidth .about]
	} else {
		set size [winfo reqheight .about]
	}
	wm geometry .about ${size}x$size
	update
	#set wx [expr ([winfo screenwidth  .about]-[winfo reqwidth  .about])/2]
	#set wy [expr ([winfo screenheight .about]-[winfo reqheight .about])/2]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  .about])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight .about])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry .about +$wx+$wy

	wm deiconify .about
	grab set .about
	focus .about
	bind .about <ButtonRelease-1> {destroy .about}
	bind .about <Escape> {destroy .about}
}

proc showtextfile {w title file args} {
	set txt "File \"$file\" not found!"
	catch {set txt [string map {"\014" ""} [readFile $file auto]]}
	showtext $w $title [encoding convertfrom identity $txt] $args
}

proc showtext {w title txt args} {
	set winname [string trim $w "."]
	set searchstringvar ::main::${winname}_searchstring
	set searchcasevar   ::main::${winname}_searchcase
	set searchregexpvar ::main::${winname}_searchregexp

	if {! [winfo exists $w]} {
		if {! [info exists $searchregexpvar]} {
			set $searchstringvar ""
			set $searchcasevar 0
			set $searchregexpvar "-exact"
		}

		toplevel $w
		wm withdraw $w
		wm title $w $title
		bind $w <Escape> "set ::config(${w}_geometry) \[wm geometry $w\]; destroy $w"
		wm protocol $w WM_DELETE_WINDOW "set ::config(${w}_geometry) \[wm geometry $w\]; destroy $w"


		frame $w.a
		pack  $w.a -side top -fill both -expand 1 -padx 4 -pady 4
		text  $w.a.txt -xscrollcommand "$w.a.xs set" \
				-yscrollcommand "$w.a.ys set" \
				-font $::config(fixed_width_font) -wrap none
		scrollbar $w.a.ys -command "$w.a.txt yview"
		scrollbar $w.a.xs -command "$w.a.txt xview" -orient horizontal
		grid  $w.a.txt -column 0 -row 0 -sticky nesw
		grid  $w.a.ys  -column 1 -row 0 -sticky ns
		grid  $w.a.xs  -column 0 -row 1 -sticky ew
		grid columnconf $w.a 0 -weight 1
		grid rowconf    $w.a 0 -weight 1

		frame $w.b
		pack  $w.b -side top -pady 2 -fill x
		frame  $w.b.s -bd 2 -relief groove
		pack   $w.b.s -side left -padx 4 -fill x -expand 1
		label  $w.b.s.l -text "Find"
		pack   $w.b.s.l -side left
		entry  $w.b.s.e -textvariable $searchstringvar
		pack   $w.b.s.e -side left -fill x -padx 2 -expand 1
		bind   $w.b.s.e <Return> "::main::searchtext $w -forwards"
		bind   $w.b.s.e <3> "tk_popup $w.searcheditmenu \[winfo pointerx .\] \[winfo pointery .\] 3"
		if {! [winfo exists $w.searcheditmenu]} {
			menu $w.searcheditmenu -tearoff 0
			$w.searcheditmenu add command -label "Copy"  -accelerator "Ctrl+C" -command "clipboard clear; clipboard append \$$searchstringvar"
			$w.searcheditmenu add command -label "Paste" -accelerator "Ctrl+V" -command "if \[catch {set $searchstringvar \[clipboard get\]}\] {bell} else {$w.b.s.e icursor end}"
			$w.searcheditmenu add separator
			$w.searcheditmenu add command -label "Clear" -accelerator "Ctrl+K" -command "set $searchstringvar {}"
		}
		if {$::tcl_platform(platform) == "windows"} {set minwidth 2} else {set minwidth 1}
		button $w.b.s.b1 -text "<" -width $minwidth -command "::main::searchtext $w -backwards"
		pack   $w.b.s.b1 -side left -padx 2
		button $w.b.s.b2 -text ">" -width $minwidth -command "::main::searchtext $w -forwards"
		pack   $w.b.s.b2 -side left -padx 2
		frame  $w.b.s.cf
		pack   $w.b.s.cf -side left
		checkbutton $w.b.s.cf.c1 -bd $::config(min_bd) -text "match case" \
				-variable $searchcasevar
		pack   $w.b.s.cf.c1 -side top -anchor w
		checkbutton $w.b.s.cf.c2 -bd $::config(min_bd) -text "regular expression" \
				-variable $searchregexpvar \
				-onvalue "-regexp" -offvalue "-exact"
		pack   $w.b.s.cf.c2 -side top -anchor w

		button $w.b.save -text Save -width 8 -command "::main::savetext $w"
		pack   $w.b.save -side left -padx 4
		button $w.b.ok -text Close -width 8 -command "set ::config(${w}_geometry) \[wm geometry $w\]; destroy $w"
		pack   $w.b.ok -side left -padx 4

		# window geometry
		if [info exists ::config(${w}_geometry)] {
			wm geometry $w $::config(${w}_geometry)
		} else {
			update
			set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
			set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
			if {$wx < 10} {set wx 10}
			if {$wy < 10} {set wy 10}
			wm geometry $w +$wx+$wy
		}

		menu $w.editmenu -tearoff 0
		$w.editmenu add command -label "Copy" -accelerator "Ctrl+C" -command "::main::showtext_copy $w"
		$w.editmenu add command -label "Select all" -accelerator "Ctrl+A" -command "::main::showtext_select_all $w"
		$w.editmenu add separator
		$w.editmenu add command -label "Clear all" -command "::main::showtext_clear_all $w"
		bind $w.a.txt <3> "tk_popup $w.editmenu \[winfo pointerx .\] \[winfo pointery .\] 0"
		bind $w <Control-a> "::main::showtext_select_all $w"

		if {[lindex $args 0] == "grab"} {
			$w.b.ok configure -command "grab set [grab current]; destroy $w"
			grab set $w
		}
	}

	$w.a.txt delete 1.0 end
	$w.a.txt insert 1.0 $txt
	$w.a.txt mark set insert 1.0

	wm deiconify $w
}
proc showtext_append {w title txt} {
	if {! [winfo exists $w]} {
		showtext $w $title $txt
	} else {
		$w.a.txt insert end $txt
	}
	$w.a.txt yview moveto 1
	# Don't display the last blank line.
	if {[string range $txt end end] == "\n"} {
		$w.a.txt delete "end - 1 chars" end
		update
		$w.a.txt insert end "\n"
	}
}
proc savetext {w} {
	# Get input filename interactively
	set initialpath [file dirname $::config(dumpinitialfile)]
	set filename [tk_getSaveFile -filetypes {{"Text files" {*.txt}}} \
		-initialdir $initialpath -parent $w \
		-title "PgcEdit: save text to file..."]
	if {$filename == ""} {return}
	if {$::tcl_platform(platform) == "windows" && \
			[string toupper [file extension $filename]] != ".TXT"} {
		set filename "${filename}.txt"
	}
	set ::config(dumpinitialfile) $filename

	saveFile $filename auto [$w.a.txt get 1.0 end]
}
proc searchtext {w direction args} {
	if {$args == ""} {
		set winname [string trim $w "."]
	} else {
		set winname $args
	}

	set searchstringvar ::main::${winname}_searchstring
	set searchcasevar   ::main::${winname}_searchcase
	set searchregexpvar ::main::${winname}_searchregexp

	set idx [$w.a.txt index insert]
	if {$direction == "-forwards"} {
		set idx "$idx + 1 chars"
	} else {
		set idx "$idx - 1 chars"
	}
	if [subst $$searchcasevar] {
		set err [catch {set found [$w.a.txt search $direction [subst $$searchregexpvar] \
				-- [subst $$searchstringvar] $idx]} errstr]
	} else {
		set err [catch {set found [$w.a.txt search $direction [subst $$searchregexpvar] \
				-nocase -- [subst $$searchstringvar] $idx]} errstr]
	}
	if {$err != 0} {
		show_warning -title "Search error" -icon error \
				-message [string map {": " ":\n"} $errstr] -parent $w
	} else {
		if {$found != ""} {
			$w.a.txt mark set insert $found
			$w.a.txt see $found
			focus $w.a.txt
		} else {
			bell
		}
	}
}
proc showtext_copy {w} {
	set txt ""
	catch {set txt [$w.a.txt get sel.first sel.last]}
	if {$txt != ""} {
		clipboard clear
		clipboard append $txt
	} else {
		bell
	}
}
proc showtext_select_all {w} {
	focus $w.a.txt
	$w.a.txt tag add sel 1.0 end
}
proc showtext_clear_all {w} {
	$w.a.txt delete 1.0 end
}

#############################################################################
## Procedure:  Create_Backup
## If the backup directory already exists, then ask the user for overwrite, or
## if the ifnotexists arg is true, don't ask and don't overwrite
proc Create_Backup {ifnotexists} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	show_msg "Creating backup..."
	set backupdir "$::dvddir/PgcEdit_backup"
	if [file exists $backupdir] {
		if $ifnotexists {
			return
		} else {
			set r [tk_messageBox -title "PgcEdit: Backup" -icon question \
				-message "Backup already exists!\nOverwrite?" -parent . \
				-type okcancel -default cancel]
			if {$r == "cancel"} return
			file delete -force -- "$backupdir/VIDEO_TS.IFO"
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				file delete -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO"
			}
			file delete -force -- $backupdir
			set dummy [file exists $backupdir]	;# wait till directory is deleted
		}
	}
	set err [catch {file mkdir $backupdir} errtxt]
	if {$err > 0} {
		errmsg $err "Error creating backup" $errtxt
	} else {
		file copy -force -- "$::dvddir/VIDEO_TS.IFO" $backupdir
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			file copy -force -- "$::dvddir/VTS_[format %02i $vts]_0.IFO" $backupdir
		}
		show_msg "Backup created OK."
	}
}

#############################################################################
## Procedure:  Restore_Backup
proc Restore_Backup {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set backupdir [file join $::dvddir PgcEdit_backup]
	if {! [file exists $backupdir]} {
		show_warning -title "PgcEdit: Backup" -icon error \
			-message "Backup not found!" -parent .
		return
	}
	show_msg "Restoring backup..."
	file copy -force -- "$backupdir/VIDEO_TS.IFO" $::dvddir
	file copy -force -- "$backupdir/VIDEO_TS.IFO" "$::dvddir/VIDEO_TS.BUP"
	for {set vts 1} {$vts <= 99} {incr vts} {
		if [file exists "$backupdir/VTS_[format %02i $vts]_0.IFO"] {
			file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" $::dvddir
			file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" \
					"$::dvddir/VTS_[format %02i $vts]_0.BUP"
		}
	}
	# are there any VOB file to restore?
	set vobfiles [glob -nocomplain -tails -directory $backupdir "*.VOB"]
	if {$vobfiles != ""} {
		#::main::dvd_modified true false
		set txt "Warning: there are VOB files in your backup folder."
		set txt "$txt\n\nThe VOB files are NOT restored automatically."
		set txt "$txt\nPlease restore them yourself."
		set txt "$txt\n\nAfter you restored them, you must re-save the DVD"
		set txt "$txt\nto update the VTS Sectors!"
		tk_messageBox -title "PgcEdit: Backup" -icon warning \
				-message $txt -parent . \
				-type ok
	}

	set r [tk_messageBox -title "PgcEdit: Backup" -icon question \
			-message "Backup restored.\nReload DVD in editor?" -parent . \
			-type okcancel -default ok]
	show_msg ""
	if {$r == "ok"} ReRead_DVD
}

#############################################################################
## Procedure:  Delete_Backup
proc Delete_Backup {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set backupdir "$::dvddir/PgcEdit_backup"
	if {! [file exists $backupdir]} {
		show_warning -title "PgcEdit: Backup" -icon error \
			-message "Backup not found!" -parent .
		return
	}
	file delete -force -- $backupdir
	show_msg "Backup deleted."
}

#############################################################################
## Procedure:  Write_DVD
## write the current DVD to disk.
## Get output filename interactively if saveas is true, else use global $::dvddir
proc Write_DVD {saveas} {

	if {! [info exists ::dvddir]} {
		bell
		return
	}

	if $saveas {
		set destdir [tk_chooseNewDirectory -initialdir $::dvddir \
			-title "Save IFOs in folder:"]
		if {$destdir == ""} {return -1}
		#set ::config(initialdir) $destdir
		set errcode [::writedvd::writedvd $destdir]
		if {$errcode == 0} {
			show_warning -title "PgcEdit: Save IFOs as" -icon warning \
					-message "IFOs saved OK!\nWARNING: VTS sectors calculated without VOBs!" \
					-parent .
		} elseif {$errcode == -1999} {
			show_warning -title "PgcEdit: Save IFOs as" -icon warning \
					-message "IFOs saved OK!\nVTS sectors ERROR!" -parent .
		}
	} else {
		set errcode [::writedvd::writedvd $::dvddir]
		if {$errcode == 0} {
			show_warning -title "PgcEdit: Save DVD" -icon info \
					-message "DVD saved OK!\nVTS sectors OK!" -parent .
			dvd_modified false false
		} elseif {$errcode == -1999} {
			show_warning -title "PgcEdit: Save DVD" -icon warning \
					-message "DVD saved OK!\nVTS sectors ERROR!" -parent .
		}
	}
	return $errcode
}

#############################################################################
## Procedure:  Read_New_DVD
## read a new DVD in memory, and init the interface.
## Get input filename interactively if args is empty.
proc Read_New_DVD {args} {

	if {$::dvd_modified && !$::config(expert_mode)} {
		set r [tk_messageBox -message "The DVD has been modified.\n\nAre you sure?" \
				-icon warning -title "PgcEdit" -type yesno -default no]
		if {$r == "no"} {return 0}
	}

	if {$args == ""} {
		set initialdir $::config(initialdir)
		while true {
			set dvddir [tk_chooseOldDirectory -initialdir $initialdir \
				-parent . -title "PgcEdit: Open DVD folder"]
			if {$dvddir == ""} return
			if {![file exists [file join $dvddir "VIDEO_TS.IFO"]] &&
					[file isdirectory [file join $dvddir "VIDEO_TS"]]} {
				set dvddir [file join $dvddir "VIDEO_TS"]
			}
			set initialdir $dvddir
			if [file exists [file join $dvddir "VIDEO_TS.IFO"]] break
			bell
		}
	} else {
		set dvddir [file normalize [lindex $args 0]]
	}

	update idletasks
	set errcode [readdvd $dvddir ::raw_sectors ::pgcs ::menus ::ttns]
	if {$errcode == 0} {
		set ::dvddir $dvddir
		set ::config(initialdir) $dvddir
		if {$::config(auto_create_backup)} {
			Create_Backup true
		}
		# init GUI
		$::widget(main_cmd_lab) configure -text ""
		$::widget(main_cmd_lab2) configure -text ""
		if {! $::scratchbook_active} {
			set ::scratchbook_active true
			scratchbook
		}
		set_window_title
		init_tree_menu

		display_pgc pgc 0 0

		# close any old information window
		destroy .streamsinfo
		destroy .dump

		show_warning -title "PgcEdit: Open DVD" -icon info -parent . \
			-message "DVD Opened OK.\n\n$::pgcs(numvts) VTS loaded."
	} else {
		set ::dvddir x;	unset ::dvddir

		set backupdir [file join $dvddir PgcEdit_backup]
		if [file exists $backupdir] {
			set r [tk_messageBox -title "PgcEdit: Open DVD" -icon question \
				-message "Error opening DVD.\nRestore backup?" -parent . \
				-type yesno -default yes]
			if {$r == "yes"} {
				file copy -force -- "$backupdir/VIDEO_TS.IFO" $dvddir
				file copy -force -- "$backupdir/VIDEO_TS.IFO" "$dvddir/VIDEO_TS.BUP"
				for {set vts 1} {$vts <= 1000} {incr vts} {
					if [file exists "$backupdir/VTS_[format %02i $vts]_0.IFO"] {
						file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" $dvddir
						file copy -force -- "$backupdir/VTS_[format %02i $vts]_0.IFO" \
								"$dvddir/VTS_[format %02i $vts]_0.BUP"
					} else {
						break
					}
				}
				Read_New_DVD $dvddir
			}
		}
	}
	return $errcode
}

proc set_window_title {} {
	if {! [info exists ::dvddir]} return

	set txt ""

	set gn 	[::utils::get_General_Name]
	if {$gn != ""} {
		set txt "$txt - \"$gn\""
	} else {
		set txt "$txt - untitled"
	}

	set id [::utils::get_provider_ID]
	if {$id != ""} {
		set txt "${txt}; provider: $id"
	}

	if {$::dvd_modified} {
		set mod "*"
	} else {
		set mod " "
	}

	wm title . "PgcEdit - $mod \"$::dvddir/\"$txt"
}

proc dvd_modified {state onlyifnotscratchbook} {
	if {$onlyifnotscratchbook && $::scratchbook_active} return

	if {$state} {
		if {! $::dvd_modified} {
			set ::dvd_modified true
			set_window_title
		}
	} else {
		if {$::dvd_modified} {
			set ::dvd_modified false
			set_window_title
		}
	}
}

#############################################################################
## Procedure:  ReRead_DVD
## re-read the current DVD in memory, and loose changes.
proc ReRead_DVD {} {
	if [info exists ::dvddir] {
		set errcode [Read_New_DVD $::dvddir]
	} else {
		set errcode -9999
		bell
	}
	return $errcode
}

#############################################################################
## Procedure:  dump_DVD
## Dump all the DVD commands in a text string, and show the text.
proc dump_DVD {} {
	if [info exists ::dvddir] {
		set txt [::disasm::dump_all_commands "DVD \"$::dvddir/\""]
		showtext ".dump" "PgcEdit: DVD Dump" $txt
	} else {
		bell
	}
}

#############################################################################
## Procedure:  init_tree_menu
## Initialize the DVD menubutton with the DVD hierarchy stored in the ::pgcs
## and ::menus global arrays.
proc init_tree_menu {} {

	set old [.mf.f1.mb cget -menu]
	if {$old != ""} {
		destroy $old
	}
	set m .mf.f1.mb.menu
	menu $m
	.mf.f1.mb configure -menu $m

	set ::main::labelslist {}
	set ::main::cmdslist {}


	# VIDEO_TS.IFO

	# FP-PGC
	set lab [pgc_label $::config(global_font) pgc 0 0]
	set labl [split [pgc_label "" pgc 0 0] ,]
	lappend ::main::labelslist $lab
	lappend ::main::cmdslist "::main::display_pgc pgc 0 0"
	$m add cascade -label "VIDEO_TS.IFO: [lindex $labl 0]" -menu $m.vts0
	menu $m.vts0
	$m.vts0 add radiobutton -variable ::main::pgcmenu_selection -value "p,0,0" \
			-label [lindex $labl 1] -command "::main::display_pgc pgc 0 0"

	# VMG Menus
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		set lulabel [lindex [split [pgc_label $::config(global_font) menu 0 $lu 1] ,] 1]
		$m.vts0 add cascade -label $lulabel -menu $m.vts0.lu$lu
		menu $m.vts0.lu$lu
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set lab [pgc_label $::config(global_font) menu 0 $lu $pgc]
			set labl [split $lab ,]
			lappend ::main::labelslist $lab
			lappend ::main::cmdslist "::main::display_pgc menu 0 $lu $pgc"
			if {$pgc > 1 && [expr $pgc % 20] == 1} {set break 1} else {set break 0}
			$m.vts0.lu$lu add radiobutton -variable ::main::pgcmenu_selection \
					-value "m,0,$lu,$pgc" -label [lindex $labl 2] -columnbreak $break \
					-command "::main::display_pgc menu 0 $lu $pgc"
		}
	}

	# VTS
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		lappend ::main::labelslist ""
		lappend ::main::cmdslist "bell"

		set vtslabel "VTS_[format %02i $vts]_0.IFO: VTS $vts"
		if {[expr $vts % 20] == 0} {set break 1} else {set break 0}
		$m add cascade -label $vtslabel -menu $m.vts$vts -columnbreak $break
		menu $m.vts$vts

		# VTS menus
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			set lulabel [lindex [split [pgc_label $::config(global_font) menu $vts $lu 1] ,] 1]
			$m.vts$vts add cascade -label $lulabel -menu $m.vts$vts.lu$lu
			menu $m.vts$vts.lu$lu
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set lab [pgc_label $::config(global_font) menu $vts $lu $pgc]
				set labl [split $lab ,]
				lappend ::main::labelslist $lab
				lappend ::main::cmdslist "::main::display_pgc menu $vts $lu $pgc"
				if {$pgc > 1 && [expr $pgc % 20] == 1} {set break 1} else {set break 0}
				$m.vts$vts.lu$lu add radiobutton -variable ::main::pgcmenu_selection \
					-value "m,$vts,$lu,$pgc" -label [lindex $labl 2] -columnbreak $break \
					-command "::main::display_pgc menu $vts $lu $pgc"
			}
		}

		# VTS Titles
		if {$::pgcs($vts,numpgcs) > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set lab [pgc_label $::config(global_font) pgc $vts $pgc]
				set labl [split $lab ,]
				lappend ::main::labelslist $lab
				lappend ::main::cmdslist "::main::display_pgc pgc $vts $pgc"
				if {$pgc > 1 && [expr $pgc % 20] == 1} {set break 1} else {set break 0}
				$m.vts$vts add radiobutton -variable ::main::pgcmenu_selection \
					-value "p,$vts,$pgc" -label [lindex $labl 1] -columnbreak $break \
					-command "::main::display_pgc pgc $vts $pgc"
			}
		}
	}
	highlight_trace_breakpoints
}

proc init_tree_menu_no_move {} {
	if [winfo exists .mf.f2.selector.a.lb] {
		set selectoroldpos [lindex [.mf.f2.selector.a.lb yview] 0]
	}
	::main::init_tree_menu
	set indexlist [split [lindex $::editor_cmds_from 1] ","]
	if {[lindex $::editor_cmds_from 0] == "pgc"} {
		::main::display_pgc pgc  [lindex $indexlist 0] [lindex $indexlist 1]
	} else {
		::main::display_pgc menu [lindex $indexlist 0] [lindex $indexlist 1] [lindex $indexlist 2]
	}
	if [winfo exists .mf.f2.selector.a.lb] {
		.mf.f2.selector.a.lb yview moveto $selectoroldpos
	}
}

#############################################################################
## Procedure:  highlight Trace breakpoints in PGC selector
proc highlight_trace_breakpoints {} {
	if {[winfo exists .mf.f2.selector.a.lb] && [info exists ::dvddir]} {
		set redlist {}
		set bluelist {}
		set purplelist {}
		if [info exists ::tr::breakpoints] {	
			if {[info exists ::tr::breakpoints(pgc,0,0,prepost)] && \
					[info exists ::tr::breakpoints(pgc,0,0,playback)]} {
				lappend purplelist 0
			} elseif [info exists ::tr::breakpoints(pgc,0,0,prepost)] {
				lappend redlist 0
			} elseif [info exists ::tr::breakpoints(pgc,0,0,playback)] {
				lappend bluelist 0
			}
			for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
				for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
					for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
						set lab [pgc_label $::config(global_font) menu $vts $lu $pgc]
						if {[info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,prepost)] && \
								[info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,playback)]} {
							lappend purplelist [lsearch $::main::labelslist $lab]
						} elseif [info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,prepost)] {
							lappend redlist [lsearch $::main::labelslist $lab]
						} elseif [info exists ::tr::breakpoints(menu,$vts,$lu,$pgc,playback)] {
							lappend bluelist [lsearch $::main::labelslist $lab]
						}
					}
				}
			}	
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
					set lab [pgc_label $::config(global_font) pgc $vts $pgc]
					if {[info exists ::tr::breakpoints(pgc,$vts,$pgc,prepost)] && \
							[info exists ::tr::breakpoints(pgc,$vts,$pgc,playback)]} {
						lappend purplelist [lsearch $::main::labelslist $lab]
					} elseif [info exists ::tr::breakpoints(pgc,$vts,$pgc,prepost)] {
						lappend redlist [lsearch $::main::labelslist $lab]
					} elseif [info exists ::tr::breakpoints(pgc,$vts,$pgc,playback)] {
						lappend bluelist [lsearch $::main::labelslist $lab]
					}
				}
			}
		}
		# set colors
		for {set i 0} {$i < [llength $::main::labelslist]} {incr i} {
			set curcolor [.mf.f2.selector.a.lb itemcget $i -foreground]
			if {$curcolor != "" && $curcolor != "black"} {
				.mf.f2.selector.a.lb itemconfigure $i -foreground black -selectforeground black
			}
		}
		foreach i $redlist {
			.mf.f2.selector.a.lb itemconfigure $i -foreground red -selectforeground red
		}
		foreach i $bluelist {
			.mf.f2.selector.a.lb itemconfigure $i -foreground blue -selectforeground blue
		}
		foreach i $purplelist {
			.mf.f2.selector.a.lb itemconfigure $i -foreground purple -selectforeground purple
		}
	}
}

#############################################################################
## Procedure:  selector
## mode arg: "show", "hide" or "toggle"
proc selector {mode labellist cmdslist} {
	set w .mf.f2.selector

	if {! [info exists $labellist]} {
		set $labellist {}
		set $cmdslist {}
	}

	switch $mode {
		"show" {
			set show 1
		}
		"hide" {
			set show 0
		}
		"toggle" {
			if [winfo exists $w.a] {
				set show 0
			} else {
				set show 1
			}
		}
	}
	if $show {
		pack  configure .mf.f2.selector -side $::config(selector_side)

		set ::main::selector_opened 1

		frame $w.a
		listbox   $w.a.lb -exportselection 0 -selectmode browse \
				-listvariable $labellist -width 0 -height $::config(selector_height) \
				-highlightcolor gray -activestyle dotbox -selectbackground gray80 \
				-selectforeground black -yscrollcommand "$w.a.ys set" 
		scrollbar $w.a.ys -command "$w.a.lb yview"
		grid  $w.a.lb -column 0 -row 0 -sticky nesw
		grid  $w.a.ys -column 1 -row 0 -sticky ns
		grid columnconf $w.a 0 -weight 1
		grid rowconf    $w.a 0 -weight 1

		bind $w.a.lb <1> "focus $w.a.lb; ::main::selector_launch_cmd_click $w $cmdslist"
		bind $w.a.lb <Double-1> "focus $w.a.lb; if \[::main::selector_launch_cmd_click $w $cmdslist\] ::main::action_Edit_PGC"
		bind $w.a.lb <Return> "::main::action_Edit_PGC"
		bind $w.a.lb <space>  "::main::selector_launch_cmd_key $w $cmdslist"
		bind $w.a.lb <Up>     "::main::selector_launch_cmd_up $w $cmdslist"
		bind $w.a.lb <Down>   "::main::selector_launch_cmd_down $w $cmdslist"
		bind $w.a.lb <Home>   "::main::selector_set_selection $w 0; ::main::selector_launch_cmd_index $w $cmdslist 0"
		bind $w.a.lb <End>    "::main::selector_set_selection $w end; ::main::selector_launch_cmd_index $w $cmdslist end"
		bind $w.a.lb <MouseWheel> "if {%D>0} {$w.a.lb yview scroll 1 units} else {$w.a.lb yview scroll -1 units}"

		::utils::selector_popup_menu_fp     .selector_popup_menu_fp
		::utils::selector_popup_menu_menus  .selector_popup_menu_menus
		::utils::selector_popup_menu_titles .selector_popup_menu_titles
		bind $w.a.lb <2> "::main::selector_middle_click $w $cmdslist"
		bind $w.a.lb <Control-1> "::main::selector_control_click $w $cmdslist"
		bind $w.a.lb <Control-Double-1> {}
		bind $w.a.lb <3> "::main::selector_context_menu $w $cmdslist"
		pack  $w.a -side top -fill both -expand 1

		highlight_trace_breakpoints
	} else {
		set ::main::selector_opened 0

		#bind . <Destroy> {}
		destroy $w.a
		#bind . <Destroy> {cleanexit 0}
		$w configure -width 1	;# don't work with 0
	}
}
proc selector_middle_click {w cmdslist} {
	focus $w.a.lb
	if [selector_launch_cmd_click $w $cmdslist] {
		set idx [$w.a.lb index @[expr [winfo pointerx .]-[winfo rootx $w.a.lb]],[expr [winfo pointery .]-[winfo rooty $w.a.lb]]]
		set domain [lindex [$w.a.lb get $idx] 0]
		#if {$domain == "VMG" } return
		#if {$domain == "VTST"} return
		::menubuttons::display_current_menu_buttons info
	}
}
proc selector_control_click {w cmdslist} {
	focus $w.a.lb

	set idx [$w.a.lb index @[expr [winfo pointerx .]-[winfo rootx $w.a.lb]],[expr [winfo pointery .]-[winfo rooty $w.a.lb]]]
	set cmd [lindex [subst $$cmdslist] [$w.a.lb index $idx]]
	if {$cmd == "bell"} {bell; return}
	catch {
		set cur_type [lindex $cmd 1]
		if {$cur_type == "menu"} {
			set index "[lindex $cmd 2],[lindex $cmd 3],[lindex $cmd 4]"
		} else {
			set index "[lindex $cmd 2],[lindex $cmd 3]"
		}	
		if {[info exists ::tr::breakpoints($cur_type,$index,prepost)] && \
				[info exists ::tr::breakpoints($cur_type,$index,playback)]} {
			unset ::tr::breakpoints($cur_type,$index,prepost)
			unset ::tr::breakpoints($cur_type,$index,playback)
		} elseif [info exists ::tr::breakpoints($cur_type,$index,prepost)] {
			unset ::tr::breakpoints($cur_type,$index,prepost)
			set ::tr::breakpoints($cur_type,$index,playback) 1
		} elseif [info exists ::tr::breakpoints($cur_type,$index,playback)] {
			set ::tr::breakpoints($cur_type,$index,prepost) 1
			set ::tr::breakpoints($cur_type,$index,playback) 1
		} else {
			set ::tr::breakpoints($cur_type,$index,prepost) 1
		}
		highlight_trace_breakpoints
	}
}
proc selector_context_menu {w cmdslist} {
	focus $w.a.lb
	selector_launch_cmd_click $w $cmdslist
	set idx [$w.a.lb index @[expr [winfo pointerx .]-[winfo rootx $w.a.lb]],[expr [winfo pointery .]-[winfo rooty $w.a.lb]]]
	set label [$w.a.lb get $idx]
	switch "[lindex $label 0]" {
		"VMG" {
			tk_popup .selector_popup_menu_fp     [winfo pointerx .] [winfo pointery .] \
					$::main::last_fp_menu_selected
		}
		"VMGM" -
		"VTSM" {
			tk_popup .selector_popup_menu_menus  [winfo pointerx .] [winfo pointery .] \
					$::main::last_menus_menu_selected
		}
		"VTST" {
			tk_popup .selector_popup_menu_titles [winfo pointerx .] [winfo pointery .] \
					$::main::last_titles_menu_selected
		}
	}
}
proc selector_set_selection {w idx} {
	$w.a.lb activate $idx
	$w.a.lb selection clear 0 end
	$w.a.lb selection set $idx
	$w.a.lb see $idx
}
proc selector_launch_cmd_index {w cmdslist index} {
	selector_set_selection $w $index
	set list [subst $$cmdslist]
	if {$index == "end"} {set index [expr [llength $list]-1]}
	eval [lindex $list $index]
}
proc selector_launch_cmd_click {w cmdslist} {
	set idx [$w.a.lb index @[expr [winfo pointerx .]-[winfo rootx $w.a.lb]],[expr [winfo pointery .]-[winfo rooty $w.a.lb]]]
	selector_set_selection $w $idx
	set cmd [lindex [subst $$cmdslist] [$w.a.lb index $idx]]
	eval $cmd
	if {$cmd == "bell"} {return false}
	return true
}
proc selector_launch_cmd_key {w cmdslist} {
	eval [lindex [subst $$cmdslist] [$w.a.lb index active]]
}
proc selector_launch_cmd_up {w cmdslist} {
	set idx [expr [$w.a.lb index active] -1]
	if {$idx >= 0} {
		if {[lindex [subst $$cmdslist] $idx] != "bell"} {
			eval [lindex [subst $$cmdslist] $idx]
		}
		incr idx
		selector_set_selection $w $idx
	}
}
proc selector_launch_cmd_down {w cmdslist} {
	set idx [expr [$w.a.lb index active] +1]
	if {$idx < [llength [subst $$cmdslist]]} {
		if {[lindex [subst $$cmdslist] $idx] != "bell"} {
			eval [lindex [subst $$cmdslist] $idx]
		}
		incr idx -1
		selector_set_selection $w $idx
	}
}

#############################################################################
## Procedure:  refresh_listbox
## refresh the listbox with the current listbox contents: re-disassemble all
## commands, regenerate line numbers...
proc init_refresh_listbox {} {
	if $::config(highlight_gotos) {
		proc refresh_listbox {} {
			set lbitem 0
			set cmdnum 0
			set gotolines(1) ""
			set gotolines(2) ""
			set gotolines(3) ""
			set sectionnum 0
			set sectionstart(1) 0
			set sectionstart(2) 1
			set sectionstart(3) 2
			set sectionstart(4) 3
			set errorlines ""
			foreach cmd $::editor_cmds_list {
				if {[string range $cmd 0 0] == {*}} {
					set cmdnum 0
					incr sectionnum
					set sectionstart($sectionnum) $lbitem
					set sectionlbnum $lbitem
					set newcmd [formatCommand "* [lindex $cmd 1]" 0]
					$::widget(main_listbox) itemconfigure $lbitem -background $::config(systembuttonface)
					$::widget(main_listbox) itemconfigure $lbitem -selectbackground gray65
				} else {
					incr cmdnum
					if {[string range $cmd 0 0] == {[}} {
						set newcmd [formatCommand [string range $cmd 1 23] $cmdnum]
					} else {
						set newcmd [formatCommand $cmd $cmdnum]
					}
					if {[string first "ILLEGAL" $newcmd] >= 0} {
						lappend errorlines $lbitem
					}
					if {[string first "Goto" $newcmd] >= 0} {
						lappend gotolines($sectionnum) [expr $sectionlbnum + 0X[string range $newcmd 22 23]]
						lappend gotolines($sectionnum) $lbitem
					}
					$::widget(main_listbox) itemconfigure $lbitem -background white
					$::widget(main_listbox) itemconfigure $lbitem -selectbackground $::config(systemhighlight)
				}
				lset ::editor_cmds_list $lbitem $newcmd
				incr lbitem
			}
			set sectionstart([expr $sectionnum+1]) $lbitem
			for {set s 1} {$s <= $sectionnum} {incr s} {
				for {set g 0} {$g < [llength $gotolines($s)]} {incr g 2} {
					set l  [lindex $gotolines($s) $g]
					set l2 [lindex $gotolines($s) [expr $g+1]]
					if {$l > $sectionstart($s) && $l < $sectionstart([expr $s+1]) && $l != $l2} {
						catch {
							$::widget(main_listbox) itemconfigure $l -background #FFFFAF
							$::widget(main_listbox) itemconfigure $l -selectbackground #BFBF84
						}
					} else {
						$::widget(main_listbox) itemconfigure $l2 -background pink
						$::widget(main_listbox) itemconfigure $l2 -selectbackground #C09090
					}
				}
			}
			foreach l $errorlines {
				$::widget(main_listbox) itemconfigure $l -background pink
				$::widget(main_listbox) itemconfigure $l -selectbackground #C09090
			}
			if {! $::scratchbook_active} {
				set domain [lindex [$::widget(main_cmd_lab) cget -text] 0]
				switch $domain {
					"VMG" {
						set illegal {
							{\[00 .[23]}
							{\[20}
							{\[30 .[358]}
							{\[[45][12]}
							{\[[89ABCD]}
						}
					}
					"VMGM" {
#						set illegal {
#							{\[20 .5}
#							{\[30 .6 .. .. .. [4C]}
#							{\[30 .[358]}
#							{\[[45][12]}
#						}
						set illegal {
							{\[20 .5}
							{\[30 .[358]}
							{\[[45][12]}
						}
					}
					"VTSM" {
						set illegal {
							{\[00 .3}
							{\[20 .5}
							{\[30 .[28]}
							{\[30 .6 .. .. [1-9][1-9] 8}
							{\[30 .6 .. .. 0[1-9] 8}
							{\[30 .6 .. .. [1-9]0 8}
							{\[[45]2}
						}
					}
					"VTST" {
						set illegal {
							{\[00 .3}
							{\[20 .1 .. .. .. .. .. 10}
							{\[30 .[26]}
							{\[[89ABCD]}
						}
					}
				}
				foreach il $illegal {
					foreach f [lsearch -all -regexp $::editor_cmds_list $il] {
						$::widget(main_listbox) itemconfigure $f -background pink
						$::widget(main_listbox) itemconfigure $f -selectbackground #C09090
					}
				}
			}
		}

	} else {

		if {[info exists ::widget(main_listbox)] && \
				[winfo exists $::widget(main_listbox)]} {
			for {set i 0} {$i < [$::widget(main_listbox) index end]} {incr i} {
				$::widget(main_listbox) itemconfigure $i -selectbackground $::config(systemhighlight)
			}
		}

		proc refresh_listbox {} {
			set cmdnum 0
			set linenum 1
			foreach cmd $::editor_cmds_list {
				switch [string range $cmd 0 0] {
					{*} {
						set linenum 1
						set newcmd [formatCommand "* [lindex $cmd 1]" 0]
						$::widget(main_listbox) itemconfigure $cmdnum -background $::config(systembuttonface)
					}
					{[} {
						set newcmd [formatCommand [string range $cmd 1 23] $linenum]
						incr linenum
						$::widget(main_listbox) itemconfigure $cmdnum -background white
					}
					default {
						set newcmd [formatCommand $cmd $linenum]
						incr linenum
						$::widget(main_listbox) itemconfigure $cmdnum -background white
					}
				}
				lset ::editor_cmds_list $cmdnum $newcmd
				incr cmdnum
			}
		}
	}
}
init_refresh_listbox

proc pgc_label {font type vts args} {
	if {$vts == 0} {
		set label "VMG"
		if {$type == "menu"} {
			set label "${label}M"
		}
	} else {
		set label "VTS"
		if {$type == "menu"} {
			set label "${label}M"
		} else {
			set label "${label}T"
		}
		set label "$label $vts"
	}
	if {$type == "pgc"} {
		set pgc $args
		if {$vts == 0} {
			return "$label , First-Play PGC"
		} else {
			set label "$label , $pgc"
			set label [padtextat $label "VTST 9 , 9999" $font]
			if [info exists ::pgcs($vts,$pgc,TTN)] {
				set label "$label  TTN $::pgcs($vts,$pgc,TTN)"
			}
			set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999" $font]

			set label "$label ([short_pbtime [string range $::pgcs($vts,$pgc,header) 4 7]])"
			#set label [padtextat $label "VTST 9 , 999 (TTN 99) (9:99:99)" $font]
			set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (9:99)  99b." $font]
			if [info exists ::pgcs($vts,$pgc,Title)] {
				set label "$label Title $::pgcs($vts,$pgc,Title)"
			}
		}
	} else {
		set lu  [lindex $args 0]
		set label "$label , LU $lu ($::menus($vts,$lu,lang))"
		set pgc [lindex $args 1]
		set label "$label , $pgc"
		set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999" $font]
		if {[info exists ::menus($vts,$lu,$pgc,numprograms)] \
				&& $::menus($vts,$lu,$pgc,numprograms) > 0} {
			set label "$label ([short_pbtime [string range $::menus($vts,$lu,$pgc,header) 4 7]])"
			if {[info exists ::menus($vts,$lu,$pgc,numbuttons)] && \
					$::buttons::butinfo($vts,errorcode) == 0} {
				set label "$label  $::menus($vts,$lu,$pgc,numbuttons)b."
			} else {
				set label "$label  ?b."
			}
		} else {
			set label "$label (dummy)"
		}
		set label [padtextat $label "VTSM 9 , LU 9 (mm) , 999 (9:99)  99b." $font]
		if [info exists ::menus($vts,$lu,$pgc,masktext)] {
			set label "$label $::menus($vts,$lu,$pgc,masktext)"
		}
	}
	return [string trimright $label " "]
}
proc padtextat {txt template font} {
	if {$font != ""} {
		set minpos [font measure $font $template]
		incr minpos -5
		while {[font measure $font $txt] < $minpos} {
			set txt "$txt "
		}
	} else {
		set txt "$txt "
	}
	return $txt
}
proc short_pbtime {bcd} {
	set t1 [format %i [expr  [get_int bcd 0 1] / 16]][format  %i [expr  [get_int bcd 0 1] & 15]]:
	set t2 [format %i [expr  [get_int bcd 1 1] / 16]][format  %i [expr  [get_int bcd 1 1] & 15]]:
	set t3 [format %i [expr  [get_int bcd 2 1] / 16]][format  %i [expr  [get_int bcd 2 1] & 15]]

	if {$t1 == "00:"} {
		set t1 ""
		if {[string range $t2 0 0] == "0"} {
			set t2 [string range $t2 1 end]
		}
	} else {
		set t1 [string trimleft $t1 "0"]
	}

	return $t1$t2$t3
}
proc pgc_sublabel {type vts args} {
	if {$type == "pgc"} {
		set pgcnum [lindex $args 0]
		if {$vts == 0} {
			set numchapters "n/a"
			set numprograms "n/a"
			set numcells    "n/a"
		} else {
			if [info exists ::pgcs($vts,$pgcnum,pttlist)] {
				set numchapters [llength $::pgcs($vts,$pgcnum,pttlist)]
			} else {
				set numchapters 0
			}
			set numprograms $::pgcs($vts,$pgcnum,numprograms)
			set numcells    $::pgcs($vts,$pgcnum,numcells)
		}
	} else {
		set lunum  [lindex $args 0]
		set pgcnum [lindex $args 1]
		set numchapters "n/a"
		set numprograms $::menus($vts,$lunum,$pgcnum,numprograms)
		set numcells    $::menus($vts,$lunum,$pgcnum,numcells)
	}
	return " Chapters: $numchapters,  Programs: $numprograms,  Cells: $numcells "
}

#############################################################################
## Procedure:  display_pgc
## Display a PGC in the listbox, and update the PGC label beside the tree menubutton
## usage:
##  display_pgc "pgc" vtsnum pgcnum
##  display_pgc "menu" vtsnum lunum pgcnum
proc display_pgc {type vts args} {
	if {$type == "pgc"} {
		set pgcnum [lindex $args 0]
		set label [pgc_label "" $type $vts $pgcnum]
		set label_selector [pgc_label $::config(global_font) $type $vts $pgcnum]
		cmds_to_listbox pgc $vts,$pgcnum $label
		set ::main::pgcmenu_selection "p,$vts,$pgcnum"
		$::widget(main_cmd_lab2) configure -text [pgc_sublabel $type $vts $pgcnum]
	} else {
		set lunum  [lindex $args 0]
		set pgcnum [lindex $args 1]
		set label [string trim [pgc_label "" $type $vts $lunum $pgcnum] "+"]
		set label_selector [string trim [pgc_label $::config(global_font) $type $vts $lunum $pgcnum] "+"]
		cmds_to_listbox menu $vts,$lunum,$pgcnum $label
		set ::main::pgcmenu_selection "m,$vts,$lunum,$pgcnum"
		$::widget(main_cmd_lab2) configure -text [pgc_sublabel $type $vts $lunum $pgcnum]
	}
	# needed for the Trace operation
	namespace eval ::tr {}
	if {$vts > 0} {
		set ::tr::last_visited_vts $vts
	} elseif {! [info exists ::tr::last_visited_vts]} {
		set ::tr::last_visited_vts 1
	}

	# refresh the PGC Selector
	set w .mf.f2.selector.a.lb
	if {[winfo exists $w]} {
		set idx [lsearch [$w get 0 end] $label_selector]
		$w selection clear 0 end
		$w selection set $idx
		$w activate $idx
		$w see $idx
	}
}

#############################################################################
## Procedure:  cmds_to_listbox
## init the global variable ::editor_cmds_list with the current commands
proc cmds_to_listbox {type index label} {

	$::widget(main_cmd_lab) configure -text $label

	if $::scratchbook_active {
		set ::scratchbook_old_selection [get_selection_indexes]
		set ::scratchbook_active false
	}

	if [info exists ::editor_cmds_from] {
		set oldtype $::editor_cmds_from
	} else {
		set oldtype ""
	}

	# remember from where the commands are coming from
	set ::editor_cmds_from [list $type $index]

	# init editor list
	switch $type {
		pgc {
			set ::editor_cmds_list $::pgcs($index)
			if {$oldtype != $::editor_cmds_from} {
				set ::editor_undo_buffer $::pgcs($index)
			}
		}
		menu {
			set ::editor_cmds_list $::menus($index)
			if {$oldtype != $::editor_cmds_from} {
				set ::editor_undo_buffer $::menus($index)
			}
		}
	}
	refresh_listbox
	set_listbox_selection 0
}

#############################################################################
## Procedure:  listbox_to_cmds
## copies the listbox contents from $::editor_cmds_list to the original part
## of the DVD $::pgcs($index) or $::menus($index)
proc listbox_to_cmds {} {
	set rawcmds {}
	foreach line $::editor_cmds_list {
		if {[string range $line 0 0] == "*"} {
			lappend rawcmds "* [lindex $line 1]"
		} else {
			lappend rawcmds "[string range $line 1 23]"
		}
	}
	if $::scratchbook_active {
		set ::scratchbook_buffer $rawcmds
	} else {
		set type  [lindex $::editor_cmds_from 0]
		set index [lindex $::editor_cmds_from 1]
		set [subst ::${type}s($index)] $rawcmds
	}
}

#############################################################################
## Procedure:  scratchbook
## Configure the listbox for the scratchbook or the normal DVD commands upon
## state of the bool $::scratchbook_active
proc scratchbook {} {
	set oldselection $::scratchbook_old_selection
	if {[lindex $oldselection 0] == "-"} {set oldselection {0 0}}
	if $::scratchbook_active {
		set ::scratchbook_old_selection [get_selection_indexes]
		set ::editor_cmds_list $::scratchbook_buffer
		set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]
	} else {
		if [info exists ::editor_cmds_from] {
			set ::scratchbook_old_selection [get_selection_indexes]
			set type  [lindex $::editor_cmds_from 0]
			set index [lindex $::editor_cmds_from 1]
			set ::editor_cmds_list [subst \$::${type}s($index)]
			set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]
		} else {
			set ::scratchbook_active true
			bell
		}
	}
	refresh_listbox
}
proc save_scratchbook {} {
	if {[llength $::scratchbook_buffer] < 2} {
		bell
		return
	}
	set filename [tk_getSaveFile -filetypes {{"scratchbook files" {*.sbk}}} \
		-initialdir "[file dirname $::config(scratchbook_initialfile)]" \
		-initialfile "[file tail $::config(scratchbook_initialfile)]" \
		-title "PgcEdit: Save scratchbook" -parent .]
	if {$filename == ""} {
		return
	}
	if {[file extension $filename] != ".sbk"} {
		set filename "${filename}.sbk"
	}
	set ::config(scratchbook_initialfile) $filename
	saveFile $filename binary $::scratchbook_buffer
}
proc open_scratchbook {} {
	set filename [tk_getOpenFile \
		-filetypes {{"scratchbook files" {*.sbk}} {"text files" {*.txt}} {"all files" {*.*}}} \
		-initialdir "[file dirname $::config(scratchbook_initialfile)]" \
		-initialfile "[file tail $::config(scratchbook_initialfile)]" \
		-title "PgcEdit: Open scratchbook" -parent .]
	if {$filename == ""} {
		return
	}
	if {[file extension $filename] != ".sbk"} {
		set filename "${filename}.sbk"
	}
	set ::config(scratchbook_initialfile) $filename
	set buf [readFile $filename binary]
	if [catch {set line0 [lindex $buf 0]}] {
		bell
		return
	}
	if {[lindex $line0 1] != "scratchbook"} {
		bell
		return
	}
	set ::scratchbook_buffer $buf
	set ::scratchbook_active true
	scratchbook
}

#############################################################################
## Procedures:  Import/Export current/all PGC commands

## TODO: Check all commands for the correct format
proc import_one_pgc {filename varname} {
	set buf [readFile $filename binary]
	set pre  [lsearch $buf "* pre"]
	set post [lsearch $buf "* post"]
	if {$pre != 0} {bell; return 0}
	if {$post < $pre} {bell; return 0}
	if {[lsearch $buf "* cell"] < $post} {bell; return 0}

	set $varname $buf
	return true
}
proc import_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]

	set filename [tk_getOpenFile \
		-filetypes {{"PGC commands files" {*.pgc}} {"text files" {*.txt}} {"all files" {*.*}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile [get_default_filename $::editor_cmds_from] -title "PgcEdit: Import PGC commands"]
	if {$filename == ""} {
		return
	}
	if {[file extension $filename] != ".pgc"} {
		set filename "${filename}.pgc"
	}
	set ::config(import_initialdir) [file dirname $filename]

	import_one_pgc $filename ::${type}s($index)
	dvd_modified true false

	set indexlst [split $index ","]
	display_pgc ${type} [lindex $indexlst 0] [lindex $indexlst 1] [lindex $indexlst 2]
}
proc import_all_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	if {!$::config(expert_mode)} {
		set txt "PgcEdit will try to (re)import all the PGCs of the current DVD."
		set txt "$txt\nThe filenames must be in the standard Import/Export PGC format."
		set txt "$txt\nIf a file does't exist, the PGC will not be imported."
		set txt "$txt\n\nWarning: Only some basic checks are done to ensure that the datas"
		set txt "$txt\nare in the correct format."
		set txt "$txt\n\nContinue?"
		set r [tk_messageBox -title "PgcEdit: Import All PGC commands" -icon question \
				-message $txt -parent . -type yesno]
		if {$r == "no"} {
			return
		}
	}
	set dirname [tk_chooseOldDirectory -parent . \
		-initialdir $::config(import_initialdir) \
		-title "PgcEdit: Import All PGC commands from folder"]
	if {$dirname == ""} {
		return
	}
	set ::config(import_initialdir) $dirname

	dvd_modified true false
	set count   0
	set skipped 0
	set errors  0

	set name [get_default_filename {::pgc 0,0}]
	set filename [file join $dirname $name]
	if [file exists $filename] {
		show_msg "Importing $name"
		if [import_one_pgc $filename ::pgcs(0,0)] {
			incr count
		} else {
			incr errors
		}
	} else {
		incr skipped
	}
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set name [get_default_filename [list "menu" "$vts,$lu,$pgc"]]
				set filename [file join $dirname $name]
				if [file exists $filename] {
					show_msg "Importing $name"
					if [import_one_pgc $filename ::menus($vts,$lu,$pgc)] {
						incr count
					} else {
						incr errors
					}
				} else {
					incr skipped
				}
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set name [get_default_filename [list "pgc" "$vts,$pgc"]]
				set filename [file join $dirname $name]
				if [file exists $filename] {
					show_msg "Importing $name"
					if [import_one_pgc $filename ::pgcs($vts,$pgc)] {
						incr count
					} else {
						incr errors
					}
				} else {
					incr skipped
				}
			}
		}
	}
	display_pgc "pgc" 0 0
	show_msg "$count PGCs imported OK.  $skipped PGCs skipped.  $errors errors."
}
proc export_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]

	set cmds  [subst \$::${type}s($index)]
	set filename [tk_getSaveFile -filetypes {{"PGC commands files" {*.pgc}}} \
		-parent . -initialdir $::config(import_initialdir) \
		-initialfile [get_default_filename $::editor_cmds_from] -title "PgcEdit: Export PGC commands"]
	if {$filename == ""} {
		return
	}
	if {[file extension $filename] != ".pgc"} {
		set filename "${filename}.pgc"
	}
	set ::config(import_initialdir) [file dirname $filename]
	saveFile $filename binary $cmds
}
proc export_all_pgc_cmds {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	set dirname [tk_chooseNewDirectory -parent . \
			-initialdir $::config(import_initialdir) \
			-title "Export All PGC commands to folder:"]
	if {$dirname == ""} {
		return
	}
	set ::config(import_initialdir) $dirname

	set count 0

	set cmds $::pgcs(0,0)
	set name [get_default_filename {::pgc 0,0}]
	show_msg "Exporting $name"
	set filename [file join $dirname $name]
	if {[saveFile $filename binary $cmds] != 0} {
		show_msg "$count PGCs exported OK.  Error exporting $name!"
		bell
		return
	}
	incr count
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set cmds $::menus($vts,$lu,$pgc)
				set name [get_default_filename [list "menu" "$vts,$lu,$pgc"]]
				show_msg "Exporting $name"
				set filename [file join $dirname $name]
				if {[saveFile $filename binary $cmds] != 0} {
					show_msg "$count PGCs exported OK.  Error exporting $name!"
					bell
					return
				}
				incr count
			}
		}
		if {$vts > 0} {
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set cmds $::pgcs($vts,$pgc)
				set name [get_default_filename [list "pgc" "$vts,$pgc"]]
				show_msg "Exporting $name"
				set filename [file join $dirname $name]
				if {[saveFile $filename binary $cmds] != 0} {
					show_msg "$count PGCs exported OK.  Error exporting $name!"
					bell
					return
				}
				incr count
			}
		}
	}
	show_msg "$count PGCs exported OK."
}
proc get_default_filename {from} {
	set type  [lindex $from 0]
	set index [lindex $from 1]
	set indexlst [split $index ","]
	if {[lindex $indexlst 0] == 0} {
		set vts "vmg"
	} else {
		set vts "vts[format "%02d" [lindex $indexlst 0]]"
	}
	if {$type == "menu"} {
		set defaultfilename "${vts}_menu[lindex $indexlst 1]_[format "%02d" [lindex $indexlst 2]].pgc"
	} else {
		if {[lindex $indexlst 0] == 0} {
			set ttn "fp"
		} else {
			set ttn "ttn[format "%02d" [lindex $indexlst 1]]"
		}
		set defaultfilename "${vts}_$ttn.pgc"
	}
}

#############################################################################
## Procedure:  set_listbox_selection
## set the selection of the listbox to the provided range
## Args: index1 ?index2? (if index2 is not provided, select only one line)
proc set_listbox_selection {idx1 args} {
	if {$args == ""} {set idx2 $idx1} else {set idx2 $args}
	if [catch {$::widget(main_listbox) activate $idx1}] {
		return 0
	}
	$::widget(main_listbox) selection clear 0 end
	$::widget(main_listbox) selection set $idx1 $idx2
	$::widget(main_listbox) see $idx2
	$::widget(main_listbox) see $idx1
	return 1
}

#############################################################################
## Procedure:  get_selection_indexes
## retreave the first and last indexes of the selection in the listbox.
## return a list with the two indexes, or {- -} if nothing is selected.
## Note: the two indexes are identical if only one command is selected.
## Do NOT check for selection of pre, post or cell line.
proc get_selection_indexes {} {
	set selection [$::widget(main_listbox) curselection]
	if {$selection == ""} {
		return [list - -]
	}
	set idx1 [lindex $selection 0]
	set idx2 [lindex $selection end]
	return [list $idx1 $idx2]
}

#############################################################################
## Procedure:  get_type
## return a list with the type (pre/post/cell/scratchbook) and line number
## of the provided index.
## Return "" if the index is out of range.
proc get_type {idx} {
	if {$idx >= [$::widget(main_listbox) index end]} {return ""}
	for {set c 0} {$c <= $idx} {incr c} {
		set cmdtext [$::widget(main_listbox) get $c]
		if {[string range $cmdtext 0 0] == "*"} {
			set type [lindex $cmdtext 1]
			set line 0
		} else {
			incr line
		}
	}
	return [list $type $line]
}

#############################################################################
## Procedure:  get_type_cmds
## return a list with all commands of the provided type (pre/post/cell/scratchbook).
proc get_type_cmds {type} {
	set ok false
	set r ""
	for {set c 0} {$c < [$::widget(main_listbox) index end]} {incr c} {
		set cmdtext [$::widget(main_listbox) get $c]
		if {[string range $cmdtext 0 0] == "*"} {
			if {[lindex $cmdtext 1] == $type} {
				set ok true
				continue
			} else {
				if $ok {return $r}
				set ok false
				continue
			}
		}
		if $ok {
			lappend r [string range $cmdtext 1 23]
		}
	}
	return $r
}

#############################################################################
## Procedure:  set_type_cmds
## replace all commands of the provided type (pre/post/cell/scratchbook).
proc set_type_cmds {type cmds} {
	set ok true
	set r ""
	for {set c 0} {$c < [$::widget(main_listbox) index end]} {incr c} {
		set cmdtext [$::widget(main_listbox) get $c]
		if {[string range $cmdtext 0 0] == "*"} {
			if {[lindex $cmdtext 1] == $type} {
				set ok false
				lappend r "* $type"
				set r [concat $r $cmds]
			} else {
				set ok true
			}
		}
		if {$ok} {
			lappend r $cmdtext
		}
	}
	set ::editor_cmds_list $r
	refresh_listbox
	return
}

#############################################################################
## Procedure:  renumber_gotos_del
## renumber the GoTos in the commands as if the commands from idx1 to idx2
## (inclusive) would be deleted.
## DO NOT actually delete the commands
proc renumber_gotos_del {cmds idx1 idx2} {
	set diff [expr $idx1 - $idx2 - 1]
	for {set idx 0} {$idx < [llength $cmds]} {incr idx} {
		if {$idx < [expr $idx1 - 1] || $idx >= $idx2} {
			set cmd [lindex $cmds $idx]
			set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
			if {$cmdchk == "001" || $cmdchk == "003"} {
				set ln [expr 0x[lindex $cmd end]]
				if {$ln > $idx2} {
					incr ln $diff
					lset cmd end [format "%02X" $ln]
					lset cmds $idx $cmd
				}
			}
		}
	}
	return $cmds
}

#############################################################################
## Procedure:  renumber_gotos_add
## renumber the GoTos in the commands as if the commands from idx1 to idx2
## (inclusive) would be added.
## The new commands must be added before calling this function
proc renumber_gotos_add {cmds idx1 idx2} {
	set diff [expr $idx2 - $idx1 + 1]
	for {set idx 0} {$idx < [llength $cmds]} {incr idx} {
		if {$idx < [expr $idx1 - 1] || $idx >= $idx2} {
			set cmd [lindex $cmds $idx]
			set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
			if {$cmdchk == "001" || $cmdchk == "003"} {
				set ln [expr 0x[lindex $cmd end]]
				if {$ln >= $idx1} {
					incr ln $diff
					lset cmd end [format "%02X" $ln]
					lset cmds $idx $cmd
				}
			}
		}
	}
	return $cmds
}

#############################################################################
## Procedure:  get_selected_cmds
## get the selected commands from the listbox
## return a list with the commands in hex form, or bell if none is selected,
## or if a pre, post or cell label is in the selection.
proc get_selected_cmds {} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return ""
	}
	set cmds {}
	for {set cidx $idx1} {$cidx <= $idx2} {incr cidx} {
		set cmdtext [$::widget(main_listbox) get $cidx]
		if {[string range $cmdtext 0 0] != {[} } {
			bell
			return ""
		}
		lappend cmds [string range $cmdtext 1 23]
	}
	return $cmds
}
#############################################################################
## Procedure:  delete_selected_cmds
## Return the original starting index of the selection, or "-" for error.
proc delete_selected_cmds {} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return "-"
	}

	if $::config(renumber_gotos) {
		set type1list [get_type $idx1]
		set type [lindex $type1list 0]
		set cmds [get_type_cmds $type]
		set renumbered [renumber_gotos_del $cmds \
				[lindex $type1list 1] [lindex [get_type $idx2] 1]]

		set_type_cmds $type $renumbered
	}
	set ::editor_cmds_list [lreplace $::editor_cmds_list $idx1 $idx2]
	return $idx1
}

#############################################################################
## Procedure:  insert_cmds
## insert commands in the listbox, just before the provided index.
proc insert_cmds {index commands} {
	if {[expr [llength $::editor_cmds_list]+[llength $commands]] > 131} {
		bell
		return -1
	}

	set formatted_commands ""
	foreach i $commands {
		lappend formatted_commands "\[$i\]"
	}
	set ::editor_cmds_list [concat \
			[lrange $::editor_cmds_list 0 [expr $index -1]] \
			$formatted_commands \
			[lrange $::editor_cmds_list $index end]]

	if $::config(renumber_gotos) {
		set type1list [get_type $index]
		set type [lindex $type1list 0]
		set typecmds [get_type_cmds $type]

		set renumbered [renumber_gotos_add $typecmds \
				[lindex $type1list 1] \
				[lindex [get_type [expr $index + [llength $commands] - 1]] 1]]

		set_type_cmds $type $renumbered
	}
	return [llength $::editor_cmds_list]
}

#############################################################################
## highlight goto line
proc highlight_goto {linenum jump} {
	set line [lindex $::editor_cmds_list $linenum]
	set cmd [string range $line 0 23]
	set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
	if {$cmdchk=="001" || $cmdchk=="003"} {
		set err 0
		set ln [expr 0X[string range $cmd 22 23]]
		if {$ln>0 && $ln<129} {
			set highln [expr $linenum + [expr $ln - [string range $line 26 28]]]
			if {![catch {set origcolor [$::widget(main_listbox) itemcget $highln -background]}]} {
				set destline [lindex $::editor_cmds_list $highln]
				if {[string range $destline 0 0]=="\["} {
					if {[expr [string range $destline 26 28]]==$ln} {
						$::widget(main_listbox) itemconfigure $highln -background lightblue
						after 333 "catch {$::widget(main_listbox) itemconfigure $highln \
								-background $origcolor}"
						if $jump {
							set_listbox_selection $linenum
							after 333 "catch {::main::set_listbox_selection $highln}"
						}
					} else {
						set err 1
					}
				} else {
					set err 1
				}
			} else {
				set err 1
			}
		} else {
			set err 1
		}
		if $err {
			$::widget(main_listbox) itemconfigure $linenum -selectbackground red
			if $jump bell
			if $::config(highlight_gotos) {
				set color #C09090
			} else {
				set color $::config(systemhighlight)
			}
			after 333 "$::widget(main_listbox) itemconfigure $linenum \
					-selectbackground $color"
		}
	}
}
#############################################################################
## highlight target PGC (if relevant)
proc highlight_jump {linenum jump} {
	if $::scratchbook_active {
		if $jump bell
		return 0
	}

	set cmd [string range [lindex $::editor_cmds_list $linenum] 0 23]
	set currenttype  [lindex $::editor_cmds_from 0]
	set currentindex [lindex $::editor_cmds_from 1]

	return [highlight_jump_cmd $currenttype $currentindex $cmd $linenum $jump true]
}
proc highlight_jump_cmd {currenttype currentindex cmd linenum jump highlight} {
	set currentvts  [lindex [split $currentindex ","] 0]
	set jumpfound false
	set targettype ""
	set targetvts -1
	set targetpgc -1
	set targetlu  "all"
	set targetsection "na"
	set targetline -1

	set linkpgc false

	set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
	switch $cmdchk {
		"201" {
			# Link*PGC
			set linkpgc true
		}
		"204" {
			# LinkPGCN
			set jumpfound true
			set targettype $currenttype
			set targetvts $currentvts
			if {$targettype == "menu"} {
				set targetlu [lindex [split $currentindex ","] 1]
			}
			set targetpgc [expr 0X[string range $cmd 19 20][string range $cmd 22 23]]
		}
		"205" {
			# Link_PTT
			set targettype "vts"
			if {$currentvts == 0} {
				set targetvts -1
			} else {
				set targetvts $currentvts
				set currentpgc [lindex [split $currentindex ","] 1]
				set targetttn $::pgcs($targetvts,$currentpgc,TTN)
				set targetptt [expr 0x[string range $cmd 20 20][string range $cmd 22 23] & 0x03FF]
				set jumpfound false
				for {set pgc 1} {$pgc <= $::pgcs($targetvts,numpgcs)} {incr pgc} {
					if {$::pgcs($targetvts,$pgc,TTN) == $targetttn} {
						foreach j $::pgcs($targetvts,$pgc,pttlist) {
							if {[lindex $j 1] == $targetptt} {
								set jumpfound true
								set targetpgc $pgc
								break
							}
						}
					}
					if $jumpfound break
				}
			}
		}
		"302" {
			# JumpTT
			set jumpfound true
			set title [expr 0x[string range $cmd 16 17]]
			if {$title>0 && $title<100} {
				for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
					for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
						if {[info exists ::pgcs($vts,$pgc,Title)] && \
								$::pgcs($vts,$pgc,Title) == $title} {
							set targettype "vts"
							set targetvts $vts
							set targetpgc $pgc
						}
					}
				}
			}
		}
		"303" -
		"305" {
			# JumpVTS_TT and JumpVTS_PTT
			set targettype "vts"
			if {$currentvts == 0} {
				set targetvts -1
			} else {
				set targetvts $currentvts
				set targetttn [expr 0x[string range $cmd 16 17]]
				if [info exists ::ttns($targetvts,$targetttn)] {
					set jumpfound true
					set targetpgc $::ttns($targetvts,$targetttn)
					if {$cmdchk == "305"} {
						# JumpVTS_PTT must check if the PTT
						# is in the first PGC of a multi-PGC title
						set targetptt [expr 0x[string range $cmd 8 8][string range $cmd 10 11]]
						set jumpfound false
						for {set pgc 1} {$pgc <= $::pgcs($targetvts,numpgcs)} {incr pgc} {
							if {$::pgcs($targetvts,$pgc,TTN) == $targetttn} {
								foreach j $::pgcs($targetvts,$pgc,pttlist) {
									if {[lindex $j 1] == $targetptt} {
										set jumpfound true
										set targetpgc $pgc
										break
									}
								}
							}
							if $jumpfound break
						}
					}
				}
			}
		}
		"306" -
		"308" {
			# JumpSS and CallSS
			set jumpfound true
			set targettype "menu"
			set jumptype [string range $cmd 16 16]
			switch $jumptype {
				"0" { # FP
					set targettype "pgc"
					set targetvts 0
					set targetpgc 0
				}
				"4" { # VMGM-menu
					if [info exists ::menus(0,1,numpgcs)] {
						for {set pgc 1} {$pgc <= $::menus(0,1,numpgcs)} {incr pgc} {
							if {[expr $::menus(0,1,$pgc,mask) & 0x7F] == 2} {
								set targetvts 0
								set targetpgc $pgc
							}
						}
					} else {
						set targetvts 0
						set targetpgc 0
					}
				}
				"8" { # VTSM-menu
					if {$cmdchk == "306"} {
						set vts [expr 0X[string range $cmd 13 14]]
					} else {
						set vts $currentvts
					}
					if {$vts == 0} {	;# VTS=0 means Last visited VTS, or current VTS
						if {$currentvts > 0} {
							set vts $currentvts
						} else {
							set jumpfound false
						}
					}
					if $jumpfound {
						set menutype [expr 0X[string range $cmd 17 17]]
						if [info exists ::menus($vts,1,numpgcs)] {
							for {set pgc 1} {$pgc <= $::menus($vts,1,numpgcs)} {incr pgc} {
								if {[expr $::menus($vts,1,$pgc,mask) & 0x7F] == $menutype} {
									set targetvts $vts
									set targetpgc $pgc
								}
							}
						} else {
							set targetvts 0
							set targetpgc 0
						}
					}
				}
				"C" { # VMGM-pgc
					set targetvts 0
					set targetpgc [expr 0X[string range $cmd 10 11]]
				}
			}
		}
		"411" - "421" - "431" - "441" - "461" - "511" - "521" - "531" - "541" - "561" -
		"611" - "621" - "631" - "641" - "651" - "661" - "671" - "681" - "691" - "6A1" - "6B1" -
		"711" - "721" - "731" - "741" - "751" - "761" - "771" - "781" - "791" - "7A1" - "7B1" {
			# Set SPRM and Link*PGC
			set linkpgc true
		}
		"414" - "424" - "434" - "444" - "464" - "514" - "524" - "534" - "544" - "564" -
		"614" - "624" - "634" - "644" - "654" - "664" - "674" - "684" - "694" - "6A4" - "6B4" -
		"714" - "724" - "734" - "744" - "754" - "764" - "774" - "784" - "794" - "7A4" - "7B4" {
			# Set SPRM and LinkPGCN
			set jumpfound true
			set targettype $currenttype
			set targetvts $currentvts
			set targetpgc [expr 0X[string range $cmd 19 20][string range $cmd 22 23]]
		}
	}
	switch "[string range $cmd 1 1]" {
		"8" - "9" - "A" - "B" - "C" - "D" {
			# CSetCLnk...
			set linkpgc true
		}
	}
	if $linkpgc {
		set targettype $currenttype
		set targetvts  $currentvts
		if {$targettype == "menu"} {
			set currentlu  [lindex [split $currentindex ","] 1]
			set currentpgc [lindex [split $currentindex ","] 2]
			if [catch {set pgcheader $::menus($currentvts,$currentlu,$currentpgc,header)}] {
				set pgcheader [string repeat \000 162]
			}
		} else {
			set currentlu  "all"
			set currentpgc [lindex [split $currentindex ","] 1]
			if [catch {set pgcheader $::pgcs($currentvts,$currentpgc,header)}] {
				set pgcheader [string repeat \000 162]
			}
		}
		switch "[string range $cmd 22 23]" {
			"09" {
				# LinkTopPGC
				set jumpfound true
				set targetlu  $currentlu
				set targetpgc $currentpgc
				set targetsection "pre"
			}
			"0D" {
				# LinkTailPGC
				set jumpfound true
				set targetlu  $currentlu
				set targetpgc $currentpgc
				set targetsection "post"
			}
			"0A" {
				# LinkNextPGC
				set jumpfound true
				set targetpgc [get_int pgcheader 0X9C 2]
			}
			"0B" {
				# LinkPrevPGC
				set jumpfound true
				set targetpgc [get_int pgcheader 0X9E 2]
			}
			"0C" {
				# LinkGoUpPGC
				set jumpfound true
				set targetpgc [get_int pgcheader 0XA0 2]
			}
		}
	}

	set err false
	if $jumpfound {
		if {$targetpgc == -1} {
			set err true
		} else {
			if {$targetsection != "na"} {
				if {$targetsection == "pre"} {
					set targetline 0
				} else {
					set targetline -1
					for {set c 0} {$c < [$::widget(main_listbox) index end]} {incr c} {
						set cmd [$::widget(main_listbox) get $c]
						if {[lindex $cmd 1] == "post"} {
							set targetline $c
							break
						}
					}
				}
				if {$targetline != -1 && $linenum != 0} {
					set line1 [expr $targetline+1]
					set cmd1 [$::widget(main_listbox) get $line1]
					if {[string range $cmd1 0 0] == "\["} {
						set targetline $line1
					}
					if $highlight {
						set origcolor [$::widget(main_listbox) itemcget $targetline -background]
						$::widget(main_listbox) itemconfigure $targetline -background lightblue
						after 333 "catch {$::widget(main_listbox) itemconfigure $targetline \
								-background $origcolor}"
					}
					if $jump {
						if {$linenum != -1} {set_listbox_selection $linenum}
						after 333 "catch {::main::set_listbox_selection $targetline}"
					}
				}
			} else {
				if [catch {
					if {$targettype == "menu"} {
						set label [pgc_label $::config(global_font) "menu" $targetvts 1 $targetpgc]
					} else {
						set label [pgc_label $::config(global_font) "pgc" $targetvts $targetpgc]
					}
				}] {
					set label "ERROR"
				}
				set highln [lsearch $::main::labelslist $label]
				if {$highln != -1} {
					if {$targettype == "menu"} {
						if {$targetlu == "all"} {
							set start 1
							if [catch {set end $::menus($targetvts,numlus)}] {
								set end 1
							}
						} else {
							set start $targetlu
							set end   $targetlu
						}
						for {set lu $start} {$lu <= $end} {incr lu} {
							if $highlight {
								if {! [catch {set label [pgc_label $::config(global_font) "menu" $targetvts $lu $targetpgc]}]} {
									set cmd2 ""
									if {$::main::selector_opened} {
										if {! [info exists ::main::selector_shifted]} {
											set oldpos [lindex [.mf.f2.selector.a.lb yview] 0]
											.mf.f2.selector.a.lb see $highln
											if {!$jump && [lindex [.mf.f2.selector.a.lb yview] 0] != $oldpos} {
												set ::main::selector_shifted true
												set cmd2 ".mf.f2.selector.a.lb yview moveto $oldpos; \
														unset ::main::selector_shifted"
											}
										}
										.mf.f2.selector.a.lb itemconfigure $highln -background lightblue
										after 1000 ".mf.f2.selector.a.lb itemconfigure $highln \
												-background $::config(listbox_background_color); $cmd2"
									}
								}
							}
						}
						if $jump {
							display_pgc "menu" $targetvts 1 $targetpgc
						}
					} else {
						if $highlight {
							set cmd2 ""
							if {$::main::selector_opened} {
								if {! [info exists ::main::selector_shifted]} {
									set oldpos [lindex [.mf.f2.selector.a.lb yview] 0]
									.mf.f2.selector.a.lb see $highln
									if {!$jump && [lindex [.mf.f2.selector.a.lb yview] 0] != $oldpos} {
										set ::main::selector_shifted true
										set cmd2 ".mf.f2.selector.a.lb yview moveto $oldpos; \
												unset ::main::selector_shifted"
									}
								}
								.mf.f2.selector.a.lb itemconfigure $highln -background lightblue
								after 1000 ".mf.f2.selector.a.lb itemconfigure $highln \
										-background $::config(listbox_background_color); $cmd2"
							}
						}
						if $jump {
							display_pgc "pgc" $targetvts $targetpgc
						}
					}
				} else {
					set err true
				}
			}
		}
	}
	if $err {
		if $highlight {
			if {$linenum < 1} {
				bell
			} else {
				$::widget(main_listbox) itemconfigure $linenum -selectbackground red
				if $jump bell
				if $::config(highlight_gotos) {
					set color #C09090
				} else {
					set color $::config(systemhighlight)
				}
				after 333 "$::widget(main_listbox) itemconfigure $linenum -selectbackground $color"
			}
		}
		return ""
	} else {
		if {$jumpfound} {
			if {$targettype == "menus"} {set targettype "menu"}
			return 	[list $targettype $targetvts $targetlu $targetpgc]
		} else {
			return ""
		}
	}
}
#############################################################################
## highlight goto/jump/call/link line, and eventually jump to it
## command bound to mouse button 1 on a line of the main listbox
proc highlight_goto_bound {jump} {
	set linenum [.mf.f2.main.l index \
			@[expr [winfo pointerx .]-[winfo rootx .mf.f2.main.l]],[expr [winfo pointery .]-[winfo rooty .mf.f2.main.l]]]
	highlight_goto $linenum $jump
	highlight_jump $linenum $jump
}

#############################################################################
## Find procedures
proc find_in_string {expr str case regexpr} {
	if $regexpr {
		if $case {
			set found [regexp -- $expr $str]
		} else {
			set found [regexp -nocase -- $expr $str]
		}
	} else {
		if {!$case} {
			set expr [string toupper $expr]
			set str  [string toupper $str]
		}
		if {[string first $expr $str] >= 0} {
			return 1
		} else {
			return 0
		}
	}
}
proc find_in_pgc {direction} {
	# check expression sybtax
	if {$::main::searchstring == ""} {
		show_msg "Type a search string first!"
		return -1
	}
	if $::main::searchregexp {
		set err [catch {set found [regexp -- $::main::searchstring "blah blah"]} errstr]
		if {$err != 0} {
			bell
			show_msg [string map {": " ":\n"} $errstr]
			return -1
		}
	}

	if {$::scratchbook_active} {
		set cmds  $::scratchbook_buffer
	} else {
		set type  [lindex $::editor_cmds_from 0]
		set index [lindex $::editor_cmds_from 1]
		set cmds  [subst \$::${type}s($index)]
	}

	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		set idx1 0
		set idx2 0
	}
	set found 0
	if {$direction == "-forwards"} {
		set start [expr $idx2 + 1]
		set end [expr [llength $cmds] - 1]
		for {set i $start} {$i <= $end} {incr i} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $::main::searchstring \
					[::disasm::formatCommand $cmd $i] \
					$::main::searchcase $::main::searchregexp]
			if {$found != 0} break
		}
	} else {
		set start [expr $idx1 - 1]
		set end 1
		for {set i $start} {$i >= $end} {incr i -1} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $::main::searchstring \
					[::disasm::formatCommand $cmd $i] \
					$::main::searchcase $::main::searchregexp]
			if {$found != 0} break
		}
	}
	if {$found == 1} {
		set_listbox_selection $i
		highlight_goto $i false
		highlight_jump $i false
	}
	return $found
}
proc find_in_cmds {expr cmds direction case regexpr} {
	if {$direction == "-forwards"} {
		for {set i 0} {$i < [llength $cmds]} {incr i} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $::main::searchstring \
					[::disasm::formatCommand $cmd $i] \
					$::main::searchcase $::main::searchregexp]
			if {$found == 1}  {return $i}
			#if {$found == -1} {return -1}
		}
	} else {
		for {set i [expr [llength $cmds] - 1]} {$i >= 0} {incr i -1} {
			set cmd [lindex $cmds $i]
			if {[lindex $cmd 0] == "*"} continue
			set found [find_in_string $::main::searchstring \
					[::disasm::formatCommand $cmd $i] \
					$::main::searchcase $::main::searchregexp]
			if {$found == 1} {return $i}
			#if {$found == -1} {return -1}
		}
	}
	return -1
}
proc find_in_dvd {direction} {
	if {! [info exists ::dvddir] || $::scratchbook_active} {
		bell
		show_msg "Find in DVD doesn't work with the Scratchbook!"
		return -1
	}

	# check expression sybtax
	if {$::main::searchstring == ""} {
		show_msg "Type a search string first!"
		return -1
	}
	if $::main::searchregexp {
		set err [catch {set found [regexp -- $::main::searchstring "blah blah"]} errstr]
		if {$err != 0} {
			bell
			show_msg [string map {": " ":\n"} $errstr]
			return -1
		}
	}

	show_msg "Searching in DVD..."

	# try to find in the current PGC first
	set found [find_in_pgc $direction]
	if {$found != 0} {
		show_msg ""
		return 1
	}

	# build forward or backward links for each pgc
	if {$direction == "-forwards"} {
		set last ""
		for {set vts $::pgcs(numvts)} {$vts >= 0} {incr vts -1} {
			if {$vts > 0} {
				for {set pgc $::pgcs($vts,numpgcs)} {$pgc >= 1} {incr pgc -1} {
					set next(pgc,$vts,$pgc) $last
					set last "pgc $vts $pgc"
				}
			}
			for {set lu $::menus($vts,numlus)} {$lu >= 1} {incr lu -1} {
				for {set pgc $::menus($vts,$lu,numpgcs)} {$pgc >= 1} {incr pgc -1} {
					set next(menu,$vts,$lu,$pgc) $last
					set last "menu $vts $lu $pgc"
				}
			}
		}
		set next(pgc,0,0) $last
	} else {
		set next(pgc,0,0) ""
		set last "pgc 0 0"
		for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set next(menu,$vts,$lu,$pgc) $last
					set last "menu $vts $lu $pgc"
				}
			}
			if {$vts > 0} {
				for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
					set next(pgc,$vts,$pgc) $last
					set last "pgc $vts $pgc"
				}
			}
		}
	}

	# find in previous or next PGCs
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	while {$next($type,$index) != ""} {
		set lnext $next($type,$index)
		set type  [lindex $lnext 0]
		set vts   [lindex $lnext 1]
		if {$type == "menu"} {
			set index "$vts,[lindex $lnext 2],[lindex $lnext 3]"
			set cmds $::menus($index)
		} else {
			set index "$vts,[lindex $lnext 2]"
			set cmds $::pgcs($index)
		}
		set line [find_in_cmds $::main::searchstring $cmds $direction \
				$::main::searchcase $::main::searchregexp]
		if {$line >= 0} {
			if {$type == "menu"} {
				display_pgc $type $vts [lindex $lnext 2] [lindex $lnext 3]
			} else {
				display_pgc $type $vts [lindex $lnext 2]
			}
			set_listbox_selection $line
			show_msg ""
			return 1
		}
	}
	bell
	show_msg "Search string not found."
	return 0
}

#############################################################################
## Action Procedures
## procs called by the action buttons (below the listbox)
## There are no args to theses commands

proc action_Edit_Cmd {} {
	set idxs [get_selection_indexes]
	set idx [lindex $idxs 0]
	if {$idx!=[lindex $idxs 1]} {
		bell
		return
	}
	set cmd [get_selected_cmds]
	if {$cmd == ""} {
		bell
		return
	}
	set ::hexcommand [lindex $cmd 0]

	# we use this variable to know when the command editor returns
	set ::ced_return -1
	::ced::cmdeditor
	tkwait variable ::ced_return
	if {$::ced_return == 1} {
		lset ::editor_cmds_list $idx $::hexcommand
		refresh_listbox
		listbox_to_cmds
		highlight_goto $idx false
		highlight_jump $idx false
		dvd_modified true true
	}
}

proc action_Edit_PGC {} {
	if {! [info exists ::dvddir] || $::scratchbook_active} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set pgcheader    [subst \$::${type}s($index,header)]
	set numprograms  [subst \$::${type}s($index,numprograms)]
	set programmap   [subst \$::${type}s($index,programmap)]
	set numcells     [subst \$::${type}s($index,numcells)]
	set cellpbtable  [subst \$::${type}s($index,cellpbtable)]
	set cellpostable [subst \$::${type}s($index,cellpostable)]
	if {$type == "pgc" && [info exists ::${type}s($index,pttlist)]} {
		set pttlist [subst \$::${type}s($index,pttlist)]
	} else {
		set pttlist ""
	}
	set vtsnum [lindex [split $index ","] 0]

	set r [::pgced::edit_pgc $type $index $pgcheader $numprograms $programmap \
			$numcells $cellpbtable $cellpostable $pttlist]
	if {$r != ""} {
		set ::${type}s($index,header)       [lindex $r 0]
		set ::${type}s($index,cellpbtable)  [lindex $r 1]
		set ::${type}s($index,cellpostable) [lindex $r 2]
		set ::${type}s($index,programmap)	[lindex $r 3]
		set ::${type}s($index,numprograms)	[string length [lindex $r 3]]
		if {[string range [lindex $r 3] end end] == "\000"} {
			incr ::${type}s($index,numprograms)	-1
		}
		set ::${type}s($index,pttlist)		[lindex $r 4]
		init_tree_menu_no_move
		dvd_modified true true
	}
}

proc action_Undo_pgc {} {
	if $::scratchbook_active {
		if [file exists $::config(default_scratchbook_filename)] {
			set ::scratchbook_buffer [readFile $::config(default_scratchbook_filename) binary]
			set ::editor_cmds_list $::scratchbook_buffer
			refresh_listbox
			set_listbox_selection 0
		} else bell
		return
	}
	if [info exists ::editor_undo_buffer] {
		set ::editor_cmds_list $::editor_undo_buffer
		refresh_listbox
		set_listbox_selection 0
		listbox_to_cmds
		dvd_modified true true
	} else bell
}

proc action_Delete {} {
	if {[get_selected_cmds] != ""} {
		set newidx [delete_selected_cmds]
		if {$newidx != ""} {
			refresh_listbox
			set_listbox_selection $newidx
			if {[lindex [get_selection_indexes] 0] == "-"} {
				set_listbox_selection end
			}
			listbox_to_cmds
			dvd_modified true true
		}
	}
}

proc action_Duplicate {} {
	set newcmds [get_selected_cmds]
	if {$newcmds != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		if {[insert_cmds [expr $idx2+1] $newcmds] != -1} {
			refresh_listbox
			set_listbox_selection $idx1 $idx2
			listbox_to_cmds
			dvd_modified true true
		}
	}
}

proc action_Add_before {} {
	set idx [lindex [get_selection_indexes] 0]
	if {$idx != "-" && $idx != 0} {
		if {[insert_cmds $idx {{00 00 00 00 00 00 00 00}}] != -1} {
			refresh_listbox
			set_listbox_selection $idx $idx
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Add_after {} {
	set idx [lindex [get_selection_indexes] 1]
	if {$idx != "-"} {
		incr idx 1
		if {[insert_cmds $idx {{00 00 00 00 00 00 00 00}}] != -1} {
			refresh_listbox
			set_listbox_selection $idx $idx
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Jump_to_target {} {
	set idxs [get_selection_indexes]
	set idx [lindex $idxs 0]
	if {$idx!=[lindex $idxs 1]} {
		bell
		return
	}
	set cmd [get_selected_cmds]
	if {$cmd == ""} {
		bell
		return
	}
	highlight_goto $idx true
	highlight_jump $idx true
}

# Move selected commands up in the list
# (in fact, take the command that is above the selection ($cutcmd) and move it below the selection)
proc action_Move_up {} {
	if {[get_selected_cmds] != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		set cutidx [expr $idx1 - 1]
		if {$idx1 < 2 || [string range [lindex $::editor_cmds_list $cutidx] 1 1] == "*"} {
			bell
			return
		}

		set typelist  [get_type [expr $idx1-1]]
		set type      [lindex $typelist 0]
		set srccmdnum [lindex $typelist 1]
		set dstcmdnum [expr $srccmdnum + $idx2 - $idx1 + 1]
		set typecmds  [get_type_cmds $type]

		if $::config(renumber_gotos) {
			for {set c 1} {$c <= [llength $typecmds]} {incr c} {
				set cmd [lindex $typecmds [expr $c-1]]
				set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
				if {$cmdchk=="001" || $cmdchk=="003"} {
					set gotoln [expr 0X[string range $cmd 21 22]]
					if {$gotoln == $srccmdnum} {
						set gotoln $dstcmdnum
					} elseif {$gotoln > $srccmdnum && $gotoln <= $dstcmdnum} {
						incr gotoln -1
					}
					lset typecmds [expr $c-1] \
						"[string range $cmd 0 19] [format %02X $gotoln]"
				}
			}
		}

		# move cutcmd
		set idx [expr $srccmdnum-1]
		set cutcmd [lindex $typecmds $idx]
		set typecmds [lreplace $typecmds $idx $idx]
		set typecmds [linsert $typecmds [expr $dstcmdnum-1] $cutcmd]

		set_type_cmds $type $typecmds

		refresh_listbox
		set_listbox_selection [expr $idx1-1] [expr $idx2-1]
		listbox_to_cmds
		dvd_modified true true
	}
}

# Move selected commands down in the list
# (in fact, take the command that is below the selection ($cutcmd) and move it above the selection)
proc action_Move_down {} {
	if {[get_selected_cmds] != ""} {
		set idxs [get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		set cutidx [expr $idx2 + 1]
		if {$cutidx >= [expr [llength $::editor_cmds_list]] \
				|| [string range [lindex $::editor_cmds_list $cutidx] 1 1] == "*"} {
			bell
			return
		}

		set typelist  [get_type [expr $idx2+1]]
		set type      [lindex $typelist 0]
		set srccmdnum [lindex $typelist 1]
		set dstcmdnum [expr $srccmdnum - ($idx2 - $idx1) - 1]
		set typecmds  [get_type_cmds $type]

		if $::config(renumber_gotos) {
			for {set c 1} {$c <= [llength $typecmds]} {incr c} {
				set cmd [lindex $typecmds [expr $c-1]]
				set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
				if {$cmdchk=="001" || $cmdchk=="003"} {
					set gotoln [expr 0X[string range $cmd 21 22]]
					if {$gotoln == $srccmdnum} {
						set gotoln $dstcmdnum
					} elseif {$gotoln < $srccmdnum && $gotoln >= $dstcmdnum} {
						incr gotoln
					}
					lset typecmds [expr $c-1] \
						"[string range $cmd 0 19] [format %02X $gotoln]"
				}
			}
		}

		# move cutcmd
		set idx [expr $srccmdnum-1]
		set cutcmd [lindex $typecmds $idx]
		set typecmds [lreplace $typecmds $idx $idx]
		set typecmds [linsert $typecmds [expr $dstcmdnum-1] $cutcmd]

		set_type_cmds $type $typecmds

		refresh_listbox
		set_listbox_selection [expr $idx1+1] [expr $idx2+1]
		listbox_to_cmds
		dvd_modified true true
	}
}

proc action_Copy {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		if {[llength $cmds] == 1} {
			set cmds [string trim $cmds "{}"]
		}
		clipboard clear
		clipboard append -- $cmds
	}
}

proc send_to_clipboard {} {
	if $::scratchbook_active {
		bell
		return
	}
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		foreach c $cmds {
			lappend ::scratchbook_buffer $c
		}
	}
}

proc copy_as_text {} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == ""} {
		bell
	} else {
		clipboard clear
		for {set line $idx1} {$line <= $idx2} {incr line} {
			clipboard append -- "[lindex $::editor_cmds_list $line]\n"
		}
	}
}

proc action_Append {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		set oldcmds [action_Paste_get_clipboard]
		if {$oldcmds != ""} {
			clipboard clear
			clipboard append -- [concat $oldcmds $cmds]
		}
	}
}

proc action_Cut {} {
	set cmds [get_selected_cmds]
	if {$cmds != ""} {
		if {[llength $cmds] == 1} {
			set cmds [string trim $cmds "{}"]
		}
		clipboard clear
		clipboard append -- $cmds
		action_Delete
		dvd_modified true true
	}
}

proc action_Paste_get_clipboard {} {
	if {[catch {set cmds [clipboard get]}]} {
		bell
		return ""
	}
	set cmds [string trim $cmds " \t\n"]
	if {[string range $cmds 0 0] != "\{" && [string length $cmds] == 23} {
		set cmds "\{$cmds\}"
	}

	# convert text commands to internal format
	set cmds2 ""
	set idx1 [string first "\[" $cmds]
	while {$idx1 > -1} {
		set idx2 [expr $idx1 + 24]
		set idx3 [expr $idx1 + 3]
		if {[string range $cmds $idx2 $idx2] == "\]" && [string range $cmds $idx3 $idx3] == " "} {
			incr idx1
			incr idx2 -1
			lappend cmds2 [string range $cmds $idx1 $idx2]
			set cmds [string range $cmds $idx2 end]
		} else {
			incr idx1
			set cmds [string range $cmds $idx1 end]
		}
		set idx1 [string first "\[" $cmds]
	}
	if {$cmds2 != ""} {
		set cmds $cmds2
	}

	if {[string range $cmds 0 0] != "\{"} {
		bell
		return ""
	}
	return $cmds
}

proc action_Paste_before {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	if {[expr [llength $::editor_cmds_list]+[llength $cmds]] > 131} {
		bell
		return
	}
	set idx [lindex [get_selection_indexes] 0]
	if {$idx != "-" && $idx != 0} {
		if {[insert_cmds $idx $cmds] != -1} {
			refresh_listbox
			set_listbox_selection $idx [expr $idx + [llength $cmds] -1]
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

proc action_Paste_after {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	if {[expr [llength $::editor_cmds_list]+[llength $cmds]] > 131} {
		bell
		return
	}
	set idx [expr [lindex [get_selection_indexes] 1] + 1]
	if {$idx != "-"} {
		if {[insert_cmds $idx $cmds] != -1} {
			refresh_listbox
			set_listbox_selection $idx [expr $idx + [llength $cmds] -1]
			listbox_to_cmds
			dvd_modified true true
		}
	} else {
		bell
	}
}

# TODO: Rewrite this shit
proc action_Paste_replace {} {
	set cmds [action_Paste_get_clipboard]
	if {$cmds == ""} {return}
	set renumber false

	if {[get_selected_cmds] != ""} {
		if $::config(renumber_gotos) {
			set renumber true
			set ::config(renumber_gotos) false

			set sidxs [get_selection_indexes]
			set sidx1 [lindex $sidxs 0]
			set sidx2 [lindex $sidxs 1]
			set oldnumcmds [expr $sidx2 - $sidx1 + 1]
			set numnewcmds [llength $cmds]
			set diff [expr $numnewcmds - $oldnumcmds]
			if {$diff < 0} {
				set typelist [get_type $sidx2]
				set type [lindex $typelist 0]
				set lastselectedline [lindex $typelist 1]
				set newcmds [get_type_cmds $type]
				set renumbered [renumber_gotos_del $newcmds \
						[expr $lastselectedline + $diff + 1] $lastselectedline]
				set_type_cmds $type $renumbered
			}
		}
		set idx [delete_selected_cmds]
		refresh_listbox
		set_listbox_selection [expr $idx-1]
		action_Paste_after

		if $renumber {
			if {$diff > 0} {
				set typelist [get_type $idx]
				set type [lindex $typelist 0]
				set firstline [expr [lindex $typelist 1] + [llength $cmds] - $diff]
				set lastline  [expr $firstline + $diff -1]
				set newcmds [get_type_cmds $type]
				set renumbered [renumber_gotos_add $newcmds $firstline $lastline]
				# replace the pasted cmds again, to reset the GOTOs to the original ones.
				set j 0
				for {set i [expr $firstline-2]} {$i < $lastline} {incr i} {
					lset renumbered $i [lindex $cmds $j]
					incr j
				}
				set_type_cmds $type $renumbered
			}
			set ::config(renumber_gotos) true
		}
	}
}

proc action_Mod_GOTOs {increment} {
	set idxs [get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return
	}
	set modified false
	for {set idx $idx1} {$idx <= $idx2} {incr idx} {
		set cmd [string range [lindex $::editor_cmds_list $idx] 0 23]
		set cmdchk "[string range $cmd 1 2][string range $cmd 5 5]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr 0X[string range $cmd 22 23]]
			if {$cmdchk == "003" && $ln == 0} continue
			incr ln $increment
			if {$cmdchk == "003" && $ln == 0} continue
			if {$ln <= 0 || $ln > 128} {
				bell
			} else {
				lset ::editor_cmds_list $idx \
						"\[[string range $cmd 1 20] [format %02X $ln]\]"
				set modified true
			}
		}
	}
	if {$modified} {
		refresh_listbox
		listbox_to_cmds
		dvd_modified true true
	}
}

proc action_Incr_GOTOs {} {
	action_Mod_GOTOs 1
}

proc action_Decr_GOTOs {} {
	action_Mod_GOTOs -1
}

## action buttons and/or popup menu
## mode arg: "show", "hide", "toggle" or "init"
proc action_buttons {mode} {
	set cmdlist {
		{{Edit Cmd}		{Space}}
		{{Edit PGC}		{Return}}
		{{Add before}	{Control-<}}
		{{Add after}	{Control->}}
		{{Delete}		{Delete}}
		{{Duplicate}	{Control-d}}
		{{Copy}			{Control-c}}
		{{Append}		{Insert}}
		{{Cut}			{Control-x}}
		{{Paste replace} {Control-v}}
		{{Paste before}	{Shift-V}}
		{{Paste after}	{Control-Shift-V}}
		{{Decr GOTOs}	{Control--}}
		{{Incr GOTOs}	{Control-+}}
		{{Move up}		{Control-[}}
		{{Move down}	{Control-]}}
	}
	if {! [winfo exists .actionmenu]} {
		menu .actionmenu
		if {$::config(tearoff_menus)} {
			set menunum 1
		} else {
			set menunum 0
		}
		set ::main::last_menu_selected $menunum
		foreach item $cmdlist {
			set cmd [lindex $item 0]
			if {$cmd != "-"} {
				set cmdname "::main::action_[string map {" " "_"} $cmd]"
				.actionmenu add command \
						-label $cmd -command "set ::main::last_menu_selected $menunum; $cmdname" \
						-accelerator [string map {"\\" "" "Control-" "Ctrl+" "Shift-" "Shift+"} [lindex $item 1]]
				incr menunum
			}
		}
	}

	switch $mode {
		"show" {
			set show 1
		}
		"hide" {
			set show 0
		}
		"toggle" {
			if [winfo exists .mf.cf.f] {
				set show 0
			} else {
				set show 1
			}
		}
		"init" {
			if {$::config(buttons)} {
				set show 1
			} else {
				set show 0
			}
		}
	}
	if $show {
		set ::main::buttons_opened 1

		if {! [winfo exists .mf.cf.f]} {
			frame .mf.cf.f
			set col 0
			set row 0
			foreach item $cmdlist {
				set cmd [lindex $item 0]
				if {$cmd != "-"} {
					set cmdname "::main::action_[string map {" " "_"} $cmd]"
					button .mf.cf.f.b${row}_$col -text $cmd -command $cmdname -width 10 -pady 0
					grid .mf.cf.f.b${row}_$col -column $col -row $row -sticky ew -padx 1
				}
				if {$row == 0} {
					incr row
					grid columnconf .mf.cf.f $col -weight 1
				} else {
					set row 0
					incr col
				}
				grid rowconf .mf.cf.f 0 -pad 2
				grid rowconf .mf.cf.f 1 -pad 2
			}
			pack .mf.cf.f -fill x -pady 1
		}
	} else {
		set ::main::buttons_opened 0

		if [winfo exists .mf.cf.f] {
			#bind . <Destroy> {}
			destroy .mf.cf.f
			#bind . <Destroy> {cleanexit 0}
			.mf.cf configure -height 1	;# don't work with 0
		}
	}
}

#############################################################################
## Main program entry
#############################################################################

wm withdraw .
update idletasks

#bind . <Destroy> {cleanexit 0}
wm protocol . WM_DELETE_WINDOW {cleanexit 0}

namespace import ::readdvd::*
namespace import ::disasm::*

set ::scratchbook_old_selection {0 0}
set ::scratchbook_active true
set ::scratchbook_buffer [list {* scratchbook}]
if [file exists $::config(default_scratchbook_filename)] {
	set ::scratchbook_buffer [readFile $::config(default_scratchbook_filename) binary]
}
set ::dvd_modified false

#################
#
# Build GUI.
#

# the DVD commands as they appear in the listbox window
set ::editor_cmds_list {}

frame .mf
pack  .mf -fill both -expand 1 -padx 3 -pady 2
frame .mf.tb ;# toolbar container
pack  .mf.tb -fill x -anchor w
::toolbar::init_toolbar $::config(toolbar)
set ::toolbar::toolbar_opened $::config(toolbar)
frame .mf.f1
pack  .mf.f1 -fill x -pady 1 -padx 1
menubutton .mf.f1.mb -text "PGC" -pady 0 -width 12
pack  .mf.f1.mb -side left -fill y
bind  .mf.f1.mb <3> "::main::selector toggle ::main::labelslist ::main::cmdslist"
bind  .mf.f1.mb <Control-Double-2> {console show}
frame .mf.f1.lf -bd 1 -relief sunken
pack  .mf.f1.lf -side left -padx 6 -pady 1 -fill both -expand 1
label .mf.f1.lf.lab -text "" -anchor center -bd 0
pack  .mf.f1.lf.lab -side left -padx 2 -fill both -expand 1
set    ::widget(main_cmd_lab) .mf.f1.lf.lab
bind  $::widget(main_cmd_lab) <3> {::toolbar::init_toolbar toggle}
label .mf.f1.lf.lab2 -text "" -anchor center -bd 0
pack  .mf.f1.lf.lab2 -side left -padx 2 -fill y
set ::widget(main_cmd_lab2) .mf.f1.lf.lab2
checkbutton .mf.f1.sb -text "Scratchbook" -command ::main::scratchbook \
	-variable ::scratchbook_active -onvalue true -offvalue false -width 12 \
	-indicatoron false -relief sunken -offrelief raised -selectcolor pink -pady 0
pack .mf.f1.sb -side right -fill y

# main listbox
frame .mf.f2
pack  .mf.f2 -fill both -expand 1 -pady 1 -padx 1
frame .mf.f2.selector	;# placeholder for Selector
pack  .mf.f2.selector -fill both -expand 0 -side left
frame .mf.f2.main
pack  .mf.f2.main -fill both -expand 1 -side left
listbox .mf.f2.main.l -exportselection 0 -selectmode extended \
		-xscrollcommand ".mf.f2.main.xscr set" -yscrollcommand ".mf.f2.main.yscr set" \
		-listvariable ::editor_cmds_list -font $::config(fixed_width_font) \
		-activestyle none -highlightcolor gray60 -height 3
scrollbar .mf.f2.main.yscr -command ".mf.f2.main.l yview"
scrollbar .mf.f2.main.xscr -command ".mf.f2.main.l xview" -orient horizontal
grid  .mf.f2.main.l    -column 0 -row 0 -sticky nesw
grid  .mf.f2.main.yscr -column 1 -row 0 -sticky ns
grid  .mf.f2.main.xscr -column 0 -row 1 -sticky ew
grid columnconf .mf.f2.main 0 -weight 1
grid rowconf    .mf.f2.main 0 -weight 1

# action buttons
frame .mf.cf
pack  .mf.cf -fill x
action_buttons init

# status bar and Find GUI
frame  .mf.st
pack   .mf.st -fill x -side bottom
label  .mf.st.status -text "PgcEdit v $::program_version" -anchor w -bd 1 -relief groove -pady 0
pack   .mf.st.status -side left -padx 1 -pady 0 -fill both -expand 1
bind   .mf.st.status <1> {.mf.st.status configure -text ""}
bind   .mf.st.status <3> {::main::action_buttons toggle}

entry  .mf.st.e -textvariable ::main::searchstring -width 32 -font $::config(fixed_width_font)
pack   .mf.st.e -side left -padx 2
bind   .mf.st.e <Return> "::main::find_in_pgc -forwards"
bind   .mf.st.e <Control-z> {set ::main::searchstring $::config(searchstring); .mf.st.e icursor end}
bind   .mf.st.e <3> "tk_popup .mf.st.e.searcheditmenu \[winfo pointerx .\] \[winfo pointery .\] 5"
menu .mf.st.e.searcheditmenu -tearoff 0
.mf.st.e.searcheditmenu add command -label "Restore default" -accelerator "Ctrl+Z" -command {set ::main::searchstring $::config(searchstring); .mf.st.e icursor end}
.mf.st.e.searcheditmenu add separator
.mf.st.e.searcheditmenu add command -label "Copy"  -accelerator "Ctrl+C" -command {clipboard clear; clipboard append $::main::searchstring}
.mf.st.e.searcheditmenu add command -label "Paste" -accelerator "Ctrl+V" -command {if [catch {set ::main::searchstring [clipboard get]}] {bell} else {.mf.st.e icursor end}}
.mf.st.e.searcheditmenu add separator
.mf.st.e.searcheditmenu add command -label "Clear" -accelerator "Ctrl+K" -command {set ::main::searchstring ""}
set ::main::searchstring $::config(searchstring)

if {$::tcl_platform(platform) == "windows"} {set minwidth 2} else {set minwidth 1}
button .mf.st.b1 -text "<<" -width $minwidth -pady 0 -command "::main::find_in_dvd -backwards"
pack   .mf.st.b1 -side left -padx 1 -fill y
button .mf.st.b2 -text "<"  -width $minwidth -pady 0 -command "::main::find_in_pgc -backwards"
pack   .mf.st.b2 -side left -padx 1 -fill y
button .mf.st.b3 -text ">"  -width $minwidth -pady 0 -command "::main::find_in_pgc -forwards"
pack   .mf.st.b3 -side left -padx 1 -fill y
button .mf.st.b4 -text ">>" -width $minwidth -pady 0 -command "::main::find_in_dvd -forwards"
pack   .mf.st.b4 -side left -padx 1 -fill y
checkbutton .mf.st.c1 -bd $::config(min_bd) -text "case" -variable ::main::searchcase
pack   .mf.st.c1 -side left
checkbutton .mf.st.c2 -bd $::config(min_bd) -text "reg expr" -variable ::main::searchregexp
pack   .mf.st.c2 -side left

# bindings
bind .mf.f2.main.l <1> {focus .mf.f2.main.l; ::main::highlight_goto_bound false}
#bind .mf.f2.main.l <Control-1>     {.mf.f2.main.l selection clear 0 end}
bind .mf.f2.main.l <Double-1>      ::main::action_Edit_Cmd
bind .mf.f2.main.l <Control-1> {focus .mf.f2.main.l; ::main::set_listbox_selection [.mf.f2.main.l index \
			@[expr [winfo pointerx .]-[winfo rootx .mf.f2.main.l]],[expr [winfo pointery .]-[winfo rooty .mf.f2.main.l]]]; \
			::main::highlight_goto_bound true}
bind .mf.f2.main.l <2> {focus .mf.f2.main.l; ::main::set_listbox_selection [.mf.f2.main.l index \
			@[expr [winfo pointerx .]-[winfo rootx .mf.f2.main.l]],[expr [winfo pointery .]-[winfo rooty .mf.f2.main.l]]]; \
			::main::highlight_goto_bound true}
bind .mf.f2.main.l <3> {tk_popup .actionmenu [winfo pointerx .] [winfo pointery .] $::main::last_menu_selected}
bind .mf.f2.main.l <space>			::main::action_Edit_Cmd
bind .mf.f2.main.l <Return>			::main::action_Edit_PGC
bind .mf.f2.main.l <Control-z>		::main::action_Undo_pgc
bind .mf.f2.main.l <Delete>			::main::action_Delete
bind .mf.f2.main.l <BackSpace>		::main::action_Delete
bind .mf.f2.main.l <Control-d>		::main::action_Duplicate
bind .mf.f2.main.l <Control-less>	::main::action_Add_before
bind .mf.f2.main.l <Control-greater> ::main::action_Add_after
bind .mf.f2.main.l <Insert>			::main::action_Append
bind .mf.f2.main.l <Control-bracketleft> ::main::action_Move_up
bind .mf.f2.main.l <Control-bracketright> ::main::action_Move_down
bind .mf.f2.main.l <Control-c>		::main::action_Copy
bind .mf.f2.main.l <Control-x>		::main::action_Cut
bind .mf.f2.main.l <Control-v>		::main::action_Paste_replace
bind .mf.f2.main.l <V> 				::main::action_Paste_before
bind .mf.f2.main.l <Control-V>		::main::action_Paste_after
bind .mf.f2.main.l <Control-plus>	::main::action_Incr_GOTOs
bind .mf.f2.main.l <Control-minus>	::main::action_Decr_GOTOs
bind .mf.f2.main.l <Control-j>		::main::action_Jump_to_target
bind .mf.f2.main.l <Control-a> {::main::set_listbox_selection 0 end}
bind .mf.f2.main.l <Escape> {.mf.f2.main.l see 0; .mf.f2.main.l activate 0; .mf.f2.main.l selection set 0}
bind .mf.f2.main.l <Home>   {.mf.f2.main.l see 0; .mf.f2.main.l activate 0; .mf.f2.main.l selection clear 0 end; .mf.f2.main.l selection set 0}
bind .mf.f2.main.l <End>    {.mf.f2.main.l see end; .mf.f2.main.l activate end; .mf.f2.main.l selection clear 0 end; .mf.f2.main.l selection set end}
bind .mf.f2.main.l <MouseWheel> {
	if {%D > 0} {
		.mf.f2.main.l yview scroll  1 units
	} else {
		.mf.f2.main.l yview scroll -1 units
	}
}

set ::widget(main_listbox) .mf.f2.main.l
set ::config(listbox_background_color) [$::widget(main_listbox) cget -background]


# main menu
menu .mainmenu

.mainmenu add cascade -label "File" -menu .mainmenu.file
menu .mainmenu.file
.mainmenu.file add command -label "Open DVD" -command ::main::Read_New_DVD -accelerator "Ctrl+O"
bind . <Control-o> ::main::Read_New_DVD
.mainmenu.file add command -label "Save DVD" -command {::main::Write_DVD false} -accelerator "Ctrl+S"
bind . <Control-s> {::main::Write_DVD false}
.mainmenu.file add command -label "Reopen DVD" -command ::main::ReRead_DVD -accelerator "Ctrl+R"
bind . <Control-r> ::main::ReRead_DVD
.mainmenu.file add separator
.mainmenu.file add command -label "Save IFOs as" -command {::main::Write_DVD true}
.mainmenu.file add separator
.mainmenu.file add command -label "New DVD" -command {::utils::new_DVD}
.mainmenu.file add command -label "Import VTST Titles" -command {::utils::import_VTST}
.mainmenu.file add separator
.mainmenu.file add command -label "Create backup"  -command {::main::Create_Backup false} -accelerator "Ctrl+B"
bind . <Control-b> {::main::Create_Backup false}
.mainmenu.file add command -label "Restore backup" -command ::main::Restore_Backup
.mainmenu.file add command -label "Delete backup"  -command ::main::Delete_Backup
.mainmenu.file add separator
.mainmenu.file add command -label "Open scratchbook" -command ::main::open_scratchbook
.mainmenu.file add command -label "Save scratchbook" -command ::main::save_scratchbook
.mainmenu.file add separator
.mainmenu.file add command -label "Import PGC commands" -command ::main::import_pgc_cmds -accelerator "Ctrl+I"
bind . <Control-i> ::main::import_pgc_cmds
.mainmenu.file add command -label "Export PGC commands" -command ::main::export_pgc_cmds -accelerator "Ctrl+E"
bind . <Control-e> ::main::export_pgc_cmds
.mainmenu.file add command -label "Import all PGC commands" -command ::main::import_all_pgc_cmds
.mainmenu.file add command -label "Export all PGC commands" -command ::main::export_all_pgc_cmds
.mainmenu.file add separator
.mainmenu.file add command -label "Check DVD+/-R Size" -command {::writedvd::DVD_size}
if {$::tcl_platform(platform) == "windows"} {
	.mainmenu.file add command -label "Explore DVD Folder" -command {::macros::explore_dvd} -accelerator "Ctrl+F"
	bind . <Control-f> ::macros::explore_dvd
}
if {$::config(debug) > 0 && $::tcl_platform(platform) == "windows"} {
	.mainmenu.file add separator
	.mainmenu.file add command -label "Console" -command {console show}
}
.mainmenu.file add separator
.mainmenu.file add command -label "Quit" -command {cleanexit 0}
.mainmenu.file add command -label "Kill (emergency exit)" -command exit

.mainmenu add cascade -label "Edit" -menu .actionmenu
.actionmenu add separator
set itemnum [expr [.actionmenu index end]+1]
.actionmenu add command -label "Jump to target" -accelerator "Ctrl+j" \
		-command "::main::action_Jump_to_target"
incr itemnum
.actionmenu add command -label "Undo cmds changes" -accelerator "Ctrl+z" \
		-command "set ::main::last_menu_selected $itemnum; ::main::action_Undo_pgc"
incr itemnum
.actionmenu add command -label "Copy as text" \
		-command "set ::main::last_menu_selected $itemnum; ::main::copy_as_text"
incr itemnum
.actionmenu add command -label "Append to scratchbook" \
		-command "set ::main::last_menu_selected $itemnum; ::main::send_to_clipboard"


.mainmenu add cascade -label "Trace" -menu .mainmenu.trace
menu .mainmenu.trace
.mainmenu.trace add checkbutton -variable ::tr::tracemode -label "Trace mode" -command ::tr::dvd_trace -accelerator "Ctrl+T"
bind . <Control-t> ::tr::dvd_trace
.mainmenu.trace add separator
.mainmenu.trace add command -label "Toggle PGC pre/post breakpoint" -command {::tr::toggle_breakpoint prepost} -accelerator "Ctrl+Shift+T"
.mainmenu.trace add command -label "Toggle PGC playback breakpoint" -command {::tr::toggle_breakpoint playback} -accelerator "Ctrl+Alt+T"
bind . <Control-T> {::tr::toggle_breakpoint prepost}
bind . <Control-Alt-t> {::tr::toggle_breakpoint playback}
.mainmenu.trace add command -label "Clear all PGC breakpoints" -command ::tr::clear_all_breakpoints
.mainmenu.trace add separator
.mainmenu.trace add command -label "Virtual Player Setup" -command ::tr::player_setup
.mainmenu.trace add separator
.mainmenu.trace add checkbutton -label "Hide toolbars in Trace mode ?" \
		-variable ::tr::hide_gui  -onvalue true -offvalue false


.mainmenu add cascade -label "Utilities" -menu .mainmenu.utils
::utils::utils_menu .mainmenu.utils

.mainmenu add cascade -label "Info" -menu .mainmenu.info
::macros::info_menu .mainmenu.info

.mainmenu add cascade -label "Macros" -menu .mainmenu.macros
::macros::macros_menu .mainmenu.macros

.mainmenu add cascade -label "Tools" -menu .mainmenu.tools
menu .mainmenu.tools
::toolscfg::configure_tools_menu

.mainmenu add cascade -label "View" -menu .mainmenu.view
menu .mainmenu.view
.mainmenu.view add checkbutton -label "Toolbar" \
		-variable ::toolbar::toolbar_opened -onvalue true -offvalue false \
		-command {::toolbar::init_toolbar $::toolbar::toolbar_opened}
.mainmenu.view add checkbutton -label "PGC Selector" \
		-variable ::main::selector_opened \
		-command {::main::selector toggle ::main::labelslist ::main::cmdslist}
.mainmenu.view add cascade -label "PGC Selector side" -menu .mainmenu.view.selectorside
menu .mainmenu.view.selectorside
.mainmenu.view.selectorside add radiobutton -label "Left" \
		-variable ::config(selector_side) -value "left" \
		-command {pack  configure .mf.f2.selector -side left}
.mainmenu.view.selectorside add radiobutton -label "Right" \
		-variable ::config(selector_side) -value "right" \
		-command {pack  configure .mf.f2.selector -side right}
.mainmenu.view.selectorside add separator
.mainmenu.view.selectorside add radiobutton -label "Top" \
		-variable ::config(selector_side) -value "top" \
		-command {pack  configure .mf.f2.selector -side top}
.mainmenu.view.selectorside add radiobutton -label "Bottom" \
		-variable ::config(selector_side) -value "bottom" \
		-command {pack  configure .mf.f2.selector -side bottom}
.mainmenu.view.selectorside add command -label "Increase height" \
		-command {if {$::config(selector_height)<50} {incr ::config(selector_height); .mf.f2.selector.a.lb configure -height $::config(selector_height)}}
.mainmenu.view.selectorside add command -label "Decrease height" \
		-command {if {$::config(selector_height)>3} {incr ::config(selector_height) -1; .mf.f2.selector.a.lb configure -height $::config(selector_height)}}
.mainmenu.view.selectorside add command -label "Default height (8 lines)" \
		-command {set ::config(selector_height) 8; .mf.f2.selector.a.lb configure -height 8}
.mainmenu.view add checkbutton -label "Edit Buttons" \
		-variable ::main::buttons_opened \
		-command {::main::action_buttons toggle}
.mainmenu.view add separator
.mainmenu.view add command -label "Tcl Calculator/Converter" \
		-command {::calc::calculator .calc}

.mainmenu add cascade -label "Options" -menu .mainmenu.options
menu .mainmenu.options

if {$::tcl_platform(platform) == "windows" && [info exists ::freewrap::progname]} {
	.mainmenu.options add cascade -label "Install" -menu .mainmenu.options.install
	.mainmenu.options add separator

	menu .mainmenu.options.install
	.mainmenu.options.install add checkbutton \
			-label "Shortcut icon to PgcEdit in Start Menu?" \
			-variable ::config(shortcut_startmenu) \
			-command {::winshellext::create_shortcut_to_PgcEdit startmenu $::config(shortcut_startmenu)}
	.mainmenu.options.install add checkbutton \
			-label "Shortcut icon to PgcEdit on Desktop?" \
			-variable ::config(shortcut_desktop) \
			-command {::winshellext::create_shortcut_to_PgcEdit desktop $::config(shortcut_desktop)}
	.mainmenu.options.install add checkbutton \
			-label "Shortcut icon to PgcEdit in Send To context menu?" \
			-variable ::config(shortcut_sendto) \
			-command {::winshellext::create_shortcut_to_PgcEdit sendto $::config(shortcut_sendto)}
	.mainmenu.options.install add command \
			-label "Create another shortcut icon to PgcEdit" \
			-command {::winshellext::create_shortcut_to_PgcEdit anydir true}
	.mainmenu.options.install add separator
	.mainmenu.options.install add checkbutton \
			-label "Associate PgcEdit with IFO files?" \
			-variable ::config(assoicated_with_IFO) \
			-command {::winshellext::associate_PgcEdit "IFO"}
	.mainmenu.options.install add checkbutton \
			-label "Associate PgcEdit with BUP files?" \
			-variable ::config(assoicated_with_BUP) \
			-command {::winshellext::associate_PgcEdit "BUP"}
	.mainmenu.options.install add separator
	.mainmenu.options.install add command \
			-label "Uninstall PgcEdit" \
			-command {::winshellext::uninstallPgcEdit}
}

.mainmenu.options add checkbutton -label "Launch Open DVD dialog at startup?" \
		-variable ::config(auto_open_new_dvd) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Use file browser instead of folder browser for Open/Save folder operations?" \
		-variable ::config(use_file_browser) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Show toolbar at startup?" \
		-variable ::config(toolbar) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Show PGC Selector at startup?" \
		-variable ::config(selector) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Show Edit Buttons at startup?" \
		-variable ::config(buttons) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Create backup when a DVD is opened for the first time?" \
		-variable ::config(auto_create_backup) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Adjust Command Editor window size?" \
		-variable ::config(adjust_win_size) -onvalue true -offvalue false
.mainmenu.options add cascade -label "'Menu buttons' video screen size" \
		-menu .mainmenu.options.butscrnsize
	menu .mainmenu.options.butscrnsize
	.mainmenu.options.butscrnsize add radiobutton -label "Full size" \
			-variable ::config(menubuttons_screen_divisor) -value 1.0
	.mainmenu.options.butscrnsize add radiobutton -label "Medium" \
			-variable ::config(menubuttons_screen_divisor) -value 1.333333
	.mainmenu.options.butscrnsize add radiobutton -label "Small" \
			-variable ::config(menubuttons_screen_divisor) -value 2.0
.mainmenu.options add checkbutton -label "Tearoff menus?  (require restart)" \
		-variable ::config(tearoff_menus) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Highlight Errors and GOTO targets? (Don't detect ALL errors!)" \
		-variable ::config(highlight_gotos) -onvalue true -offvalue false \
		-command {::main::init_refresh_listbox ; ::main::refresh_listbox}
.mainmenu.options add checkbutton -label "Automatically renumber GOTOs upon cmds insert/delete?" \
		-variable ::config(renumber_gotos) -onvalue true -offvalue false \
		-command ::main::renumber_gotos_warning
.mainmenu.options add checkbutton -label "Include \[hex codes\] in Info / commands dumps ?" \
		-variable ::config(dump_hex_codes) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Expert mode?  (suppress most popup dialog boxes)" \
		-variable ::config(expert_mode) -onvalue true -offvalue false
.mainmenu.options add checkbutton -label "Show libdvdread warnings when parsing menu buttons?" \
		-variable ::config(show_libdvdread_warnings) -onvalue true -offvalue false
.mainmenu.options add separator
.mainmenu.options add command -label "Set current search string as default" \
		-command {set ::config(searchstring) $::main::searchstring}
.mainmenu.options add separator
.mainmenu.options add command -label "Global font  (require restart)" \
		-command {set f [::fontreq::fontreq .fr $::config(global_font) false]; \
				if {$f!={}} {set ::config(global_font) $f}}
.mainmenu.options add command -label "Fixed-width font" \
		-command {set f [::fontreq::fontreq .fr $::config(fixed_width_font) true]; \
				if {$f!={}} {set ::config(fixed_width_font) $f; \
				.mf.f2.main.l configure -font $::config(fixed_width_font); \
				catch {$::tr::wm_main_window.fr.a.txt configure -font $::config(fixed_width_font)}
				}}
.mainmenu.options add separator
.mainmenu.options add checkbutton -label "Save setups on exit?" \
		-variable ::config(save_setups_on_exit) -onvalue true -offvalue false
.mainmenu.options add command -label "Save setups now" -command ::save_user_prefs
.mainmenu.options add command -label "Restore defaults  (require restart)" \
		-command {catch {file delete [file join $::config(config_directory) ".pgcedit.cfg"]}; \
				set ::config(restore_defaults) true}
.mainmenu.options add separator
.mainmenu.options add command -label "Save current scratchbook as default" \
		-command {saveFile $::config(default_scratchbook_filename) binary $::scratchbook_buffer}

.mainmenu add cascade -label "Help" -menu .mainmenu.help
menu .mainmenu.help
.mainmenu.help add command -label "Doc (english)" -command {::main::showtextfile \
		.docen "PgcEdit Documentation" "$::config(apppath)/README.txt"}
.mainmenu.help add command -label "Doc (franais)" -command {::main::showtextfile \
		.docfr "PgcEdit Documentation" "$::config(apppath)/LISEZMOI.txt"}
.mainmenu.help add command -label "Version history" -command {::main::showtextfile \
		.hist "PgcEdit Versions History" "$::config(apppath)/HISTORY.txt"}
.mainmenu.help add command -label "GPL License" -command {::main::showtextfile .license \
		"GNU General Public License" "$::config(apppath)/COPYING.txt"}
.mainmenu.help add separator
.mainmenu.help add command -label "ISO639 language codes" -command {::main::showtextfile .langcodes \
		"ISO639 language codes" "$::config(apppath)/lib/data/lang_codes.txt"}
if {$::config(web_browser) != ""} {
	.mainmenu.help add separator
	.mainmenu.help add command -label "Homepage" \
			-command {catch {exec -- $::config(web_browser) \
					"http://home.tiscali.be/debie.roland/pgcedit/index.html" &}}
	.mainmenu.help add command -label "Mirror" \
			-command {catch {exec -- $::config(web_browser) \
					"http://www.videohelp.com/~r0lZ/pgcedit/index.html" &}}
}
.mainmenu.help add separator
.mainmenu.help add command -label "About" -command ::main::about

. configure -menu .mainmenu






#menu .maintracemenu
#
#.maintracemenu add cascade -label "File" -menu .maintracemenu.file
#menu .maintracemenu.file
#.maintracemenu.file add command -label "Check DVD+/-R Size" -command {::writedvd::DVD_size}
#if {$::tcl_platform(platform) == "windows"} {
#	.maintracemenu.file add command -label "Explore DVD Folder" -command {::macros::explore_dvd} -accelerator "Ctrl+F"
#	bind . <Control-f> ::macros::explore_dvd
#}
#if {$::config(debug) > 0 && $::tcl_platform(platform) == "windows"} {
#	.maintracemenu.file add separator
#	.maintracemenu.file add command -label "Console" -command {console show}
#}
#.maintracemenu.file add separator
#.maintracemenu.file add command -label "Quit" -command {cleanexit 0}
#.maintracemenu.file add command -label "Kill (emergency exit)" -command exit
#
#.maintracemenu add cascade -label "Trace" -menu .tracemenu
#
#.maintracemenu add cascade -label "Info" -menu .mainmenu.info
#
#.maintracemenu add cascade -label "Tools" -menu .mainmenu.tools
#
#.maintracemenu add cascade -label "View" -menu .maintracemenu.view
#menu .maintracemenu.view
#.maintracemenu.view add checkbutton -label "PGC Selector" \
#		-variable ::main::selector_opened \
#		-command {::main::selector toggle ::main::labelslist ::main::cmdslist}
#.maintracemenu.view add cascade -label "PGC Selector side" -menu .mainmenu.view.selectorside
#.maintracemenu.view add separator
#.maintracemenu.view add command -label "Tcl Calculator/Converter" \
#		-command {::calc::calculator .calc}
#
#.maintracemenu add cascade -label "Options" -menu .mainmenu.options
#
#.maintracemenu add cascade -label "Help" -menu .mainmenu.help




proc renumber_gotos_warning {} {
	if {! $::config(expert_mode) && $::config(renumber_gotos)} {
		set txt "Warning!\n\nWhen this option is turned ON, the GOTO lines are"
		set txt "$txt\nrenumbered when commands are inserted or deleted."
		set txt "$txt\nHowever, the commands that are copied to the clipboard"
		set txt "$txt\nand the new commands that are inserted in the listbox"
		set txt "$txt\nare NOT renumbered, because it is impossible to know"
		set txt "$txt\nwhere the GOTOs were originally branching to."
		set txt "$txt\n\nRule:\nOnly the commands that are present in the listbox"
		set txt "$txt\nBEFORE AND AFTER the operation are renumbered."
		set txt "$txt\n\nKeep this in mind, and carefully verify all GOTOs after"
		set txt "$txt\neach operation!"
		tk_messageBox -title "PgcEdit: Renumber GOTOs" -icon warning \
				-message $txt -parent .
	}
}

# start is Scratchbook mode
scratchbook

# open selector if requested
if $::config(selector) {
	selector show ::main::labelslist ::main::cmdslist
}

wm title . "PgcEdit v $::program_version"
update
if [catch {wm geometry . $::config(main_win_geometry)}] {
	wm geometry . 700x400
}
wm deiconify .
update idletasks

# Get input filename from command line (if any)
if {$argc > 0} {
	# allow the main window to open and refresh before loading the DVD
	update idletasks
	set ::pause 0
	after 500 {set ::pause 1}
	tkwait variable ::pause
	unset ::pause

	set filename [lindex $argv 0]
	if [file isdirectory $filename] {
		Read_New_DVD [file normalize $filename]
	} else {
		Read_New_DVD [file normalize [file dirname $filename]]
	}
} else {
	if $::config(auto_open_new_dvd) Read_New_DVD
}



} ;# end namespace ::main
